--disable_warnings
--disable_query_log
--disable_result_log
--source ../t/test_init.inc
--enable_result_log
--enable_query_log
--enable_warnings
--let $MASTER_1_COMMENT_2_1_BACKUP= $MASTER_1_COMMENT_2_1
let $MASTER_1_COMMENT_2_1=
  COMMENT='database "auto_test_remote", table "tbl_a", srv "s_2_1", aim "0"';
let $MASTER_1_AUTO_INCREMENT_2_1=
  AUTO_INCREMENT=20;
let $MASTER_1_AUTO_INCREMENT1=
  ALTER TABLE tbl_a AUTO_INCREMENT=30;
let $MASTER_1_AUTO_INCREMENT2=
  ALTER TABLE tbl_a AUTO_INCREMENT=10;
let $CHILD2_1_CHARSET_AUTO_INCREMENT=
  AUTO_INCREMENT=20;
--let $CHILD2_1_DROP_TABLES_BACKUP= $CHILD2_1_DROP_TABLES
let $CHILD2_1_DROP_TABLES=
  DROP TABLE IF EXISTS tbl_a;
--let $CHILD2_1_CREATE_TABLES_BACKUP= $CHILD2_1_CREATE_TABLES
let $CHILD2_1_CREATE_TABLES=
  CREATE TABLE tbl_a (
    col_a INT NOT NULL AUTO_INCREMENT,
    col_b VARCHAR(20) DEFAULT 'def',
    col_c INT NOT NULL DEFAULT 10,
    PRIMARY KEY(col_a)
  ) $CHILD2_1_ENGINE $CHILD2_1_CHARSET_AUTO_INCREMENT $CHILD2_1_CHARSET;
--let $CHILD2_1_SELECT_TABLES_BACKUP= $CHILD2_1_SELECT_TABLES
let $CHILD2_1_SELECT_TABLES=
  SELECT col_a, col_b, col_c FROM tbl_a ORDER BY col_a;
let $CHILD2_1_SELECT_ARGUMENT1=
  SELECT argument FROM mysql.general_log WHERE argument LIKE '%select %';
--let $OUTPUT_CHILD_GROUP2_BACKUP= $OUTPUT_CHILD_GROUP2
--let $OUTPUT_CHILD_GROUP2= 1
--let $USE_GENERAL_LOG_BACKUP= $USE_GENERAL_LOG
--let $USE_GENERAL_LOG= 1
