/**
 * Autogenerated by Thrift Compiler (0.8.0)
 *
 * DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
 *  @generated
 */
#ifndef cassandra_CONSTANTS_H
#define cassandra_CONSTANTS_H

#include "cassandra_types.h"

namespace org { namespace apache { namespace cassandra {

class cassandraConstants {
 public:
  cassandraConstants();

//  std::string VERSION;
  char*  cassandra_const_VERSION;
};

extern const cassandraConstants g_cassandra_constants;

}}} // namespace

#endif
