tbl_opt_ai : MySQL:65901 (AUTO_INCREMENT option on InnoDB table is ignored if added before autoinc column)
delete_low_prio : InnoDB does not use table-level locking
insert_high_prio : InnoDB does not use table-level locking
insert_low_prio : InnoDB does not use table-level locking
select_high_prio : InnoDB does not use table-level locking
update_low_prio : InnoDB does not use table-level locking
insert_delayed    : MDEV-12880 - INSERT DELAYED is not detected as inapplicable to a table under lock
lock_concurrent   : MDEV-12882 - Assertion failure
tbl_opt_index_dir : INDEX DIRECTORY option is not supported anymore
