##
##  Tests that require FB/MySQL specific features for which there are
##  no plans to port them to MariaDB
##
rpl_no_unique_check_on_lag :  unique_check_lag_threshold is not available in MariaDB
rpl_no_unique_check_on_lag_mts : unique_check_lag_threshold is not available in MariaDB
consistent_snapshot_mixed_engines : Tests START TRANSACTION WITH CONSISTENT $ENGINE_NAME SNAPSHOT
rpl_skip_trx_api_binlog_format : requires @@rpl_skip_tx_api
rpl_ddl_high_priority : DDL commands with HIGH_PRIORITY syntax are not in MariaDB
rpl_gtid_rocksdb_sys_header : MariaDB doesn't support printing "RocksDB: Last MySQL Gtid UUID" into server stderr on startup
singledelete_idempotent_recovery: MariaDB doesn't support --slave-use-idempotent-for-recovery
rpl_mts_dependency_unique_key_conflicts: MariaDB doesn't support --slave-use-idempotent-for-recovery
rpl_missing_columns_sk_update : Uses log_column_names=ON feature which is only present in FB/MySQL
optimize_myrocks_replace_into: requires @@enable_blind_replace support.
rpl_gtid_crash_safe_optimized: requires slave_gtid_info=optimized

##
##  Tests that do not fit MariaDB's test environment (Functional tests only,
##  can't have stress tests)
##
rpl_rocksdb_stress_crash : Stress test

##
## Tests that are disabled for other reasons 
##

multiclient_2pc : Didn't try with MariaDB, yet
rpl_crash_safe_wal_corrupt : Didn't try with MariaDB, yet
rpl_gtid_crash_safe : Didn't try with MariaDB, yet
rpl_gtid_crash_safe_wal_corrupt : Didn't try with MariaDB, yet
rpl_rocksdb_snapshot : Didn't try with MariaDB, yet
rpl_rocksdb_snapshot_without_gtid : Didn't try with MariaDB, yet
rpl_rocksdb_slave_gtid_info_optimized: requires slave-gtid-info=optimized which is an FB/MySQL-only feature
rocksdb_slave_check_before_image_consistency: requires slave_check_before_image_consistency feature
