#define MSG_ACCESS_VIOLATN  "Violation accès mémoire"
#define MSG_ACT_ALLOC_FAIL  "PlugInitLang: Erreur d'allocation du bloc Activity"
#define MSG_ADDVAL_ERROR    "Erreur %d dans AddValue"
#define MSG_ADD_BAD_TYPE    "Ajout d'une valeur de type %s non conforme dans un tableau %s"
#define MSG_ADD_NULL_DOM    "Ajout de la chaîne %s à un domaine nul"
#define MSG_ADPOS_IN_DICTP  "ADPOS au travail dans User_Dictp"
#define MSG_AFTER           " après: "
#define MSG_ALG_CHOICE_AUTO "Le choix du meilleur algorithme est automatique"
#define MSG_ALG_CHOICE_BAD  "Choix d'algorithme invalide, remis à AUTO"
#define MSG_ALG_CHOICE_QRY  "Utilise l'algorithme 'Query'"
#define MSG_ALG_CURLY_BRK   "Le choix de l'algorithme dépend des accolades externes"
#define MSG_ALLOC_ERROR     "Erreur d'allocation de %s"
#define MSG_ALL_DELETED     "Toutes les lignes enlevées en %.2lf sec"
#define MSG_ALTER_DB_ERR    "Impossible de déterminer la base de données à modifier"
#define MSG_AMBIG_COL_QUAL  "Qualificateur ambigu %s pour la colonne %s"
#define MSG_AMBIG_CORREL    "Select %s.* corrélation ambigue"
#define MSG_AMBIG_SPEC_COL  "Colonne spéciale ambiguë %s"
#define MSG_ANSWER_TYPE     "Réponse de type"
#define MSG_API_CONF_ERROR  "Erreur SQL: API_CONFORMANCE"
#define MSG_APPL_ACCESSIBLE "Application %s accessible"
#define MSG_APPL_ACTIVE     "Application %s encore active"
#define MSG_APPL_BAD_SAVE   "Application %s partiellement sauvegardée"
#define MSG_APPL_CREATED    "Application %s crée"
#define MSG_APPL_IS_ACTIVE  "Application déjà active"
#define MSG_APPL_NOT_INIT   "Application non initialisée"
#define MSG_APPL_NOT_LOADED "Application non chargée"
#define MSG_APPL_QUIT       "Fin de l'application %s"
#define MSG_APPL_SAVED      "Application %s sauvegardée"
#define MSG_APP_STILL_ACTIV "Application du langage %s encore active (non libérable)"
#define MSG_AREAFILE_NOTFND "Fichier Area introuvable"
#define MSG_ARGS_SYNTAX_ERR "?SetArgs erreur de syntaxe: %s inattendu après %s"
#define MSG_ARG_ALREADY_SET "Argument %d déjà alloué"
#define MSG_ARG_NOT_AN_ATTR "L'argument n'est pas un attribut (type %d erroné)"
#define MSG_ARG_OUT_CONTEXT "Argument de type @ utilisé hors contexte"
#define MSG_ARG_OUT_RANGE   "Argument de phrase valant %d hors limite"
#define MSG_ARG_PTR_NOSEM   "Argument valant %d pointe sur un noeud sans Sem"
#define MSG_ARG_PTR_NOSEMS  "Argument valant %d pointe sur un noeud sans sémantique"
#define MSG_ARG_REF_LOOP    "?Bouclage entre références croisées des arguments"
#define MSG_ARG_TWO_CONST   "Le 2ème argument de %s doit être constant"
#define MSG_ARRAY_ALLOC_ERR "Erreur d'allocation mémoire dans ARRAY"
#define MSG_ARRAY_BNDS_EXCD "Hors limite de tableau"
#define MSG_ARRAY_ERROR     "Erreur de fonctionnement k=%d n=%d"
#define MSG_ATTRIBUTE_ERROR "Erreur règle %u attribut %s: "
#define MSG_ATT_NOT_CASE    "Mauvaise valeur %d pour attribut (pas une CaseValue)"
#define MSG_ATT_POSCODE_BIG "Code attribut %d trop grand (max=%d)"
#define MSG_AVGLEN_ERROR    "avglen doit être entre %d et %d"
#define MSG_BAD_AGGREG_FUNC "Fonction aggrégée %d non supportée"
#define MSG_BAD_ARGTYPES    "Argument de type invalide pour %s"
#define MSG_BAD_ARGUMENTS   "Argument not attachés pour %s"
#define MSG_BAD_ARG_NUM     "Nombre d'arguments invalide %d"
#define MSG_BAD_ARG_TYPE    "Type d'argument %d invalide"
#define MSG_BAD_ARRAY_OPER  "Les tableaux doivent utiliser l'opérateur IN"
#define MSG_BAD_ARRAY_TYPE  "Type=%d invalide pour un tableau"
#define MSG_BAD_ARRAY_VAL   "Les tableaux doivent avoir le même nombre de valeurs"
#define MSG_BAD_BIN_FMT     "Format invalide %c pour la colonne BIN %s"
#define MSG_BAD_BLK_ESTIM   "Nombre de blocs supérieur à l'estimation"
#define MSG_BAD_BLK_SIZE    "Taille du bloc %d non conforme"
#define MSG_BAD_BYTE_NUM    "Le nombre d'octets écrits est faux"
#define MSG_BAD_BYTE_READ   "Le nombre d'octets lus est faux"
#define MSG_BAD_CARDINALITY "Appel invalide de Cardinality pour une table multiple"
#define MSG_BAD_CASE_SPEC   "Min/Maj: spécification %c incorrecte, recommencez: "
#define MSG_BAD_CHAR_SPEC   "Spécification '%s' invalide pour caractère"
#define MSG_BAD_CHECK_TYPE  "Sous-type %d invalide pour CheckColumn"
#define MSG_BAD_CHECK_VAL   "Valeur pour Check invalide '%s'"
#define MSG_BAD_COLCRT_ARG  "COLCRT: Arg invalide (type=%hd, domain=%hd)"
#define MSG_BAD_COLDEF_TYPE "Coldefs: type illégal %d"
#define MSG_BAD_COLIST_ITEM "Elément invalide dans une Colist"
#define MSG_BAD_COLIST_TYPE "Mauvais type=%d pour une Colist"
#define MSG_BAD_COLSIZE     "Colsize %d trop petit pour cette base de données"
#define MSG_BAD_COL_ENTRY   "Entrée invalide pour la colonne %s"
#define MSG_BAD_COL_FORMAT  "Type de formattage %d invalide pour une colonne"
#define MSG_BAD_COL_IN_FILT "Colonne incorrecte dans un filtre"
#define MSG_BAD_COL_QUALIF  "Qualificateur invalide %s pour la colonne %s"
#define MSG_BAD_COL_TYPE    "Type invalide %s pour la colonne %s"
#define MSG_BAD_COL_XPATH   "Xpath invalide colonne %s de la table HTML %s"
#define MSG_BAD_COMPARE_OP  "Opérateur de comparaison %d invalide"
#define MSG_BAD_CONST_TYPE  "Type=%d invalide pour une constante"
#define MSG_BAD_CONV_TYPE   "Convertion de type invalide %d"
#define MSG_BAD_CORREL      "Select %s.* corrélation absente"
#define MSG_BAD_DATETIME    "Valeur date/temps invalide"
#define MSG_BAD_DATE_OPER   "Opérateur de date inattendu %d"
#define MSG_BAD_DBF_FILE    "Le fichier DBF %s est altéré"
#define MSG_BAD_DBF_REC     "Fichier DBF %s altéré enregistrement %d"
#define MSG_BAD_DBF_TYPE    "Type DBF %c non supporté"
#define MSG_BAD_DEF_ARG     "Argument invalide pour INDEXDEF (type=%hd, domain=%hd)"
#define MSG_BAD_DEF_READ    "EOF inattendue en lecture différée"
#define MSG_BAD_DEF_TYPE    "Type de colonne invalide"
#define MSG_BAD_DIRECTORY   "Répertoire invalide %s: %s"
#define MSG_BAD_DIST_JN_FIL "Filtre de jointure distincte invalide"
#define MSG_BAD_DIST_JOIN   "Spécification invalide de jointure distincte"
#define MSG_BAD_DOM_COL_DEF "Définition de colonnes invalide pour un domaine"
#define MSG_BAD_DOM_VALUE   "La valeur %d n'appartient pas au domaine"
#define MSG_BAD_EDIT_INIT   "Coparm: édition %s initialisée improprement"
#define MSG_BAD_EVAL_TYPE   "Fonction scalaire de type=%d invalide"
#define MSG_BAD_EXEC_MODE   "Mode d'exécution invalide '%s'"
#define MSG_BAD_EXP_ARGTYPE "Argument de type %d invalide pour une expression"
#define MSG_BAD_EXP_OPER    "Opérateur=%d invalide pour expression"
#define MSG_BAD_FETCH_RC    "Code retour inattendu de Fetch %d"
#define MSG_BAD_FIELD_FMT   "Format de champ invalide %c pour %s"
#define MSG_BAD_FIELD_RANK  "Rang %d invalide pour la colonne %s"
#define MSG_BAD_FIELD_TYPE  "Mauvais type de champ %s"
#define MSG_BAD_FILE_HANDLE "Handle de fichier invalide: %s"
#define MSG_BAD_FILE_LIST   "La section liste de fichiers est erronée"
#define MSG_BAD_FILTER      "Mauvais filtre: Opc=%d B_T=%d %d Type=%d %d"
#define MSG_BAD_FILTER_CONV "Conversion filtre incorrecte, B_T=%d,%d"
#define MSG_BAD_FILTER_LINK "Opérateur de chaînage illégal %d"
#define MSG_BAD_FILTER_OP   "Opérateur de filtre invalide %d"
#define MSG_BAD_FILTEST_OP  "Opérateur invalide %d %d pour FilTest"
#define MSG_BAD_FLD_FORMAT  "Format invalide pour le champs %d de %s"
#define MSG_BAD_FLD_LENGTH  "Champs %s trop long (%s --> %d) ligne %d de %s"
#define MSG_BAD_FLOAT_CONV  "Convertion invalide d'un tableau flottant"
#define MSG_BAD_FPARM_NEXT  "Coparm: FPARM avec Next non nul"
#define MSG_BAD_FREQ_SET    "Spécification erronnée de Freq pour la colonne %s"
#define MSG_BAD_FUNC_ARG    "Funcarg de type %d non implémenté"
#define MSG_BAD_FUNC_ARGTYP "Mauvais type d'argument=%d pour une fonction"
#define MSG_BAD_FUNC_MODE   "%s: mode invalide %d"
#define MSG_BAD_GENRE       "Genre est invalide"
#define MSG_BAD_GETVIEW_RET "GetView: type de retour %d invalide"
#define MSG_BAD_HANDLE_VAL  "Valeur Handle invalide"
#define MSG_BAD_HAV_FILTER  "Filtre Having sur une requête non groupée"
#define MSG_BAD_HAV_FILTYPE "Filtre invalide pour clause Having"
#define MSG_BAD_HEADER      "Fichier %s: bloc en-tête altéré"
#define MSG_BAD_HEADER_VAL  "Valeur invalide pour Header"
#define MSG_BAD_HEAD_END    "Lecture fin d'en-tête impossible"
#define MSG_BAD_INDEX_COL   "Colonne %s invalide pour index %s"
#define MSG_BAD_INDEX_DEF   "Définition invalide pour index %s"
#define MSG_BAD_INDEX_FILE  "Fichier index %s corrompu"
#define MSG_BAD_INDEX_PART  "Définition colonne invalide pour index %s"
#define MSG_BAD_INPUT       "Entrée incorrecte"
#define MSG_BAD_IN_ARGTYPE  "Argument de type invalide pour l'opérateur IN"
#define MSG_BAD_IN_ENDING   "Erreur: fin de chaîne IN invalide"
#define MSG_BAD_IN_STRING   "La chaîne IN commence ou finie par des caractères invalides %c ... %c"
#define MSG_BAD_JCOL_TYPE   "Erreur logique JCT: disparité des types colonnes"
#define MSG_BAD_JOIN_EXP    "Expression invalide pour une jointure"
#define MSG_BAD_JOIN_FILTER "Filtre de jointure invalide"
#define MSG_BAD_JOIN_OP     "Opérateur de joint invalide %d"
#define MSG_BAD_LANG_SIZE   "Le fichier langage a une mauvaise taille %d"
#define MSG_BAD_LINEFLD_FMT "Format invalide ligne %d champs %d de %s"
#define MSG_BAD_LINE_LEN    "Longueur ligne non égale à Lrecl"
#define MSG_BAD_LIST_TYPE   "Type de liste invalide %d"
#define MSG_BAD_LOCALE      "Locale invalide %s"
#define MSG_BAD_LOCDFON_ARG "Mauvais paramètre pour LOCDFON"
#define MSG_BAD_LOCNODE_USE "Usage inattendu de LOCNODE"
#define MSG_BAD_LRECL       "Disparité lrecl table/fichier (%d,%hd)"
#define MSG_BAD_MAX_HAVING  "MAXTMP trop petit pour Having"
#define MSG_BAD_MAX_NREC    "MaxRec=%d ne correspond pas à MaxBlk=%d Nrec=%d"
#define MSG_BAD_MAX_PARAM   "Mauvais paramètres pour spécifier une valeur maximum"
#define MSG_BAD_MAX_SETTING "Mauvaise valeur '%c' pour max"
#define MSG_BAD_MERGE_TYPE  "Le type %d ne pas être intercallé"
#define MSG_BAD_NODE_TYPE   "Type noeud erroné pour la table"
#define MSG_BAD_OFFSET_VAL  "Nul offset invalide pour une table CSV"
#define MSG_BAD_OPEN_MODE   "Mode d'ouverture invalide %d"
#define MSG_BAD_OPERATOR    "Opérateur invalide %s"
#define MSG_BAD_ORDER_MODE  "Mode de tri %c invalide"
#define MSG_BAD_ORDER_TYPE  "Tri sur objet de type=%d invalide"
#define MSG_BAD_OUTER_JOIN  "Jointure externe invalide sur table enfant"
#define MSG_BAD_PAD_ARGTYP  "Argument de type invalide pour Pad ou Justify"
#define MSG_BAD_PARAMETERS  "%.8s: Mauvais paramètres"
#define MSG_BAD_PARAM_TYPE  "%.8s: Paramètre de type=%d invalide"
#define MSG_BAD_PARM_COUNT  "Nombre de paramètres incohérent"
#define MSG_BAD_PHASE_NUM   "Numéro de phrase %d hors limite"
#define MSG_BAD_PHRASE_NB   "numéro de phrase hors limite %d rc=%d\n"
#define MSG_BAD_POS_CODE    "POS_code invalide %d"
#define MSG_BAD_POS_TYPE    "Type de POS_code invalide %d"
#define MSG_BAD_PROJNUM     "Mauvais projnum %d pour la colonne %s"
#define MSG_BAD_QUERY_OPEN  "Mode invalide %d pour l'ouverture d'une requête"
#define MSG_BAD_QUERY_TYPE  "Type de requête %d invalide pour %s"
#define MSG_BAD_QUOTE_FIELD "Quote manquante dans %s champs %d ligne %d"
#define MSG_BAD_READ_NUMBER "Mauvais nombre %d de valeurs lues dans %s"
#define MSG_BAD_RECFM       "Recfm type %d invalide pour DOSCOL"
#define MSG_BAD_RECFM_VAL   "Valeur invalide %d de Recfm"
#define MSG_BAD_RESULT_TYPE "Mauvais type de résultat %d pour %s"
#define MSG_BAD_RETURN_TYPE "Type de retour %d incorrect"
#define MSG_BAD_ROW_VALIST  "Liste de valeurs invalide pour ROW"
#define MSG_BAD_ROW_VALNB   "Nombre de valeurs inégal dans la liste"
#define MSG_BAD_SCF_ARGTYPE "Argument %d de type=%s invalide pour %s"
#define MSG_BAD_SEM_DOMAIN  "Domain .%d invalide"
#define MSG_BAD_SETTINGS    "Certaines spécifications sont incompatibles avec le type de la table"
#define MSG_BAD_SET_CASE    "La casse d'un tableau ne peut pas passer de non respect à respecter"
#define MSG_BAD_SET_STRING  "SetValue: appel invalide pour STRING"
#define MSG_BAD_SET_TYPE    "Set type %hd invalide"
#define MSG_BAD_SPECIAL_CMD "Commande spéciale invalide"
#define MSG_BAD_SPECIAL_COL "Colonne spéciale invalide %s"
#define MSG_BAD_SPEC_COLUMN "Colonne spéciale invalide pour ce type de table"
#define MSG_BAD_SQL_PARAM   "Paramètre SQL invalide pour FindColblk"
#define MSG_BAD_SUBLST_TYPE "Coparm: type %d de sous-liste invalide"
#define MSG_BAD_SUBSEL_IN_X "Sub-select invalide pour une expression"
#define MSG_BAD_SUBSEL_TYPE "Type %d invalide retourné de Sub-Select"
#define MSG_BAD_SUB_RESULT  "Résultat indéfini de fonction Sub-Select"
#define MSG_BAD_SUB_SELECT  "Sub-select invalide comme argument de fonction"
#define MSG_BAD_TABLE_LINE  "Ligne '%s' illégale ou tronquée dans la section Tables"
#define MSG_BAD_TABLE_LIST  "Table %s absente de la liste des tables"
#define MSG_BAD_TABLE_TYPE  "Type invalide %s pour la table %s"
#define MSG_BAD_TEST_TYPE   "BlockTest sur tableau: types dépareillés %s %s"
#define MSG_BAD_TRIM_ARGTYP "Argument de type invalide pour Trim"
#define MSG_BAD_TYPE_FOR_IN "Types d'argument incompatibles pour la fonction IN"
#define MSG_BAD_TYPE_FOR_S  "Type incorrecte %d pour %s(%d)"
#define MSG_BAD_TYPE_LIKE   "Type(%d)= %d invalide pour LIKE"
#define MSG_BAD_UPD_COR     "Le qualificateur %s de la colonne %s ne se refère pas à la table mise à jour %s"
#define MSG_BAD_USERBLK_LEN "Mauvaise longueur à l'écriture du bloc utilisateur"
#define MSG_BAD_USETEMP     "Usetemp invalide '%s'"
#define MSG_BAD_USETEMP_VAL "Valeur pour Usetemp invalide %d"
#define MSG_BAD_VALBLK_INDX "Valeur hors limites de l'index du bloc de valeurs"
#define MSG_BAD_VALBLK_TYPE "Type=%d invalide pour un bloc de valeurs"
#define MSG_BAD_VALNODE     "Type %d invalide pour le noeud valeur colonne %s"
#define MSG_BAD_VALUE_TYPE  "Type de valeur invalide %d"
#define MSG_BAD_VAL_UPDATE  "Impossible de déterminer quelle valeur %s doit être mise à jour"
#define MSG_BAD_VIEW_OPEN   "Mode invalide %d pour l'ouverture d'une View"
#define MSG_BAD_XMODE_VAL   "Mode d'exécution %d invalide"
#define MSG_BAD_XOBJ_TYPE   "Mauvais type de Xobject %d"
#define MSG_BAS_NS_LIST     "Format invalide de la liste des espace-noms"
#define MSG_BIN_F_TOO_LONG  "Valeur trop longue pour le champ %s (%d --> %d)"
#define MSG_BIN_MODE_FAIL   "Echec mode binaire: %s"
#define MSG_BLKTYPLEN_MISM  "Disparité types/longueurs de bloc dans SetValue"
#define MSG_BLK_IS_NULL     "Blk est nul"
#define MSG_BLOCK_NO_MATCH  "Bloc non correspondant"
#define MSG_BREAKPOINT      "Point de contrôle"
#define MSG_BUFF_TOO_SMALL  "GetColData: Buffer trop petit"
#define MSG_BUFSIZE_ERROR   "Erreur en recherchant la taille du buffer"
#define MSG_BUILDING_GROUPS "Formation des groupes"
#define MSG_BUILD_DIST_GRPS "Formation des groupes distinctes"
#define MSG_BUILD_INDEX     "Construction index %s sur %s"
#define MSG_BXP_NULL        "Bxp nul dans PUTFON"
#define MSG_CANNOT_OPEN     "Ouverture impossible de %s"
#define MSG_CD_ONE_STEP     "Count Distinct doit être exécuté en une seule étape"
#define MSG_CD_ORDER_ERROR  "Erreur de tri dans Count Distinct"
#define MSG_CHECKING_ROWS   "Test des lignes à mettre à jour"
#define MSG_CHECK_LEVEL     "Niveau de vérification fixé à %u"
#define MSG_CHSIZE_ERROR    "Erreur dans chsize: %s"
#define MSG_CLN_NOT_IN_JOIN "La colonne C%d n'est pas dans le join"
#define MSG_CNTDIS_COL_LOST "Colonne du Count Distinct perdue"
#define MSG_COLIST_BAD_TYPE "Type=%d invalide pour Colist"
#define MSG_COLNAM_TOO_LONG "Nom de colonne trop long"
#define MSG_COLSEC_TOO_BIG  "Section colonne trop grande, table %s (%d)"
#define MSG_COLS_REDUCED    " (réduit par Maxcol)"
#define MSG_COLUMN_ERROR    "Erreur de colonne"
#define MSG_COLUMN_MISMATCH "Colonne %s dépareillée"
#define MSG_COLUMN_NOT_KEY  "La colonne jointe R%d.%s n'est pas une clé"
#define MSG_COL_ALLOC_ERR   "Allocation impossible du noeud colonne"
#define MSG_COL_ALLOC_ERROR "Erreur d'allocation mémoire pour la colonne %d"
#define MSG_COL_HAS_NO_DEF  "La colonne %s n'est pas définie"
#define MSG_COL_INVAL_TABLE "La colonne %s.%s n'existe pas dans la table %s alias %s"
#define MSG_COL_ISNOT_TABLE "La colonne %s n'est pas dans la table %s"
#define MSG_COL_NB_MISM     "Le nombre de colonnes ne correspond pas"
#define MSG_COL_NOTIN_GRPBY "La colonne %s n'est pas dans la liste de Group By"
#define MSG_COL_NOTIN_TABLE "La colonne %s n'est dans aucune table"
#define MSG_COL_NOTIN_UPDT  "%s n'appartient pas à la table mise à jour %s"
#define MSG_COL_NOT_CODED   "La colonne %s n'est pas codifiée"
#define MSG_COL_NOT_EXIST   "La colonne %s n'existe pas dans %s"
#define MSG_COL_NOT_FOUND   "La colonne %s n'est pas dans la table %s"
#define MSG_COL_NOT_IN_DB   "La colonne %s de la table %s n'est pas dans la base de données"
#define MSG_COL_NOT_IN_JOIN "La colonne %s n'est pas dans le join"
#define MSG_COL_NOT_SORTED  "La colonne %s de la table %s n'est pas triée"
#define MSG_COL_NUM_MISM    "Disparité du nombre de colonnes"
#define MSG_COL_USED_TWICE  "Colonne %s utilisée deux fois ???"
#define MSG_COMPUTE_ERROR   "Erreur dans Compute, op=%d"
#define MSG_COMPUTE_NIY     "Compute non implémenté pour TOKEN"
#define MSG_COMPUTING       "Calculs en cours"
#define MSG_COMPUTING_DIST  "Comptage des valeurs distinctes"
#define MSG_COMPUTING_FUNC  "Calcul de(s) fonction(s)"
#define MSG_COM_ERROR       "Erreur Com"
#define MSG_CONCAT_SUBNODE  "Concaténation de sous-noeuds impossible"
#define MSG_CONNECTED       "Connecté"
#define MSG_CONNECT_CANCEL  "Connection interrompue par l'utilisateur"
#define MSG_CONNECT_ERROR   "Erreur %d se connectant à %s"
#define MSG_CONN_CLOSED     "%s(%d) fermée"
#define MSG_CONN_CREATED    "Connexion %s crée"
#define MSG_CONN_DROPPED    "Connexion %s supprimée"
#define MSG_CONN_OPEN       "%s(%d) ouverte (%s)"
#define MSG_CONN_SUC_OPEN   "%s(%d) ouverte avec succès"
#define MSG_CONTROL_C_EXIT  "Exit par Ctrl-C"
#define MSG_COPY_BAD_PHASE  "Copie de liste invalide en phase %d"
#define MSG_COPY_INV_TYPE   "Coparm: type non supporté %d"
#define MSG_CORREL_NO_QRY   "Les sous-requêtes corrélées ne peuvent pas être de type QRY"
#define MSG_CREATED_PLUGDB  " Créé par PlugDB %s "
#define MSG_CURSOR_SET      "Curseur remis à %d"
#define MSG_DATABASE_ACTIVE "Base de données %s activée"
#define MSG_DATABASE_LOADED "Base de données %s chargée"
#define MSG_DATA_IS_NULL    "ExecSpecialCmd: data est NULL"
#define MSG_DATA_MISALIGN   "Mauvais alignement pour ce type de données"
#define MSG_DBASE_FILE      "Fichier dBASE dbf: "
#define MSG_DB_ALREADY_DEF  "Base de données %s déjà définie"
#define MSG_DB_ALTERED      "Base de données modifiée"
#define MSG_DB_CREATED      "Base de données %s créée"
#define MSG_DB_NOT_SPEC     "Base de données non spécifiée"
#define MSG_DB_REMOVED      "Base de données %s retirée de la liste"
#define MSG_DB_SORT_ERROR   "Erreur de tri DB"
#define MSG_DB_STOPPED      "Arrêt de la base de données %s"
#define MSG_DEBUG_NOT_ACTIV "Mode Debug inactif"
#define MSG_DEBUG_SET_INV   "Invalide pour Debug: %c"
#define MSG_DEF_ALLOC_ERROR "Erreur d'allocation de la classe DEF %s"
#define MSG_DELETING_ROWS   "Suppression des lignes"
#define MSG_DEL_FILE_ERR    "Erreur à l'effacement de %s"
#define MSG_DEL_READ_ERROR  "Delete: erreur en lecture req=%d len=%d"
#define MSG_DEL_WRITE_ERROR "Delete: erreur en écriture: %s"
#define MSG_DEPREC_FLAG     "Option Flag périmée, utiliser Coltype"
#define MSG_DICTIONARY      "Dictionnaire "
#define MSG_DIRECT_VARTOK   "Accès direct aux règles du Variable Token non implémenté"
#define MSG_DISCONNECTED    "Déconnecté"
#define MSG_DISTINCT_ERROR  "Plus d'un élément fonctionel DISTINCT"
#define MSG_DISTINCT_ROWS   "Sélection des lignes distinctes"
#define MSG_DISTINCT_VALUES "Extraction des valeurs distinctes"
#define MSG_DIS_NOHEAD_JOIN "Jointure distincte sur une table non en tête"
#define MSG_DLL_LOAD_ERROR  "Erreur %d au chargement du module %s"
#define MSG_DOMAIN_EMPTY    "Le domaine %s est vide"
#define MSG_DOMAIN_ERROR    "Colonne %s: disparité domaine(%s)/valeur(%s)"
#define MSG_DOMAIN_FULL     "Le domaine %s est plein (max=%d)"
#define MSG_DOM_FILE_ERROR  "Fichier domain %s introuvable"
#define MSG_DOM_NOT_SUPP    "MS-DOM non supporté par cette version"
#define MSG_DOM_OPEN_ERROR  "Erreur d'ouverture du domaine: %s"
#define MSG_DOM_READ_ERROR  "Erreur %d en lecture de domaine: %s"
#define MSG_DOM_READ_ONLY   "La table domaine %s est en lecture seulement"
#define MSG_DOM_WRITE_ERROR "Erreur %d en écriture de domaine: %s"
#define MSG_DONE            "Effectué, rc=%d"
#define MSG_DOSALMEM_NOMEM  "Erreur d'allocation, pas assez de mémoire"
#define MSG_DROP_DB_ERR     "Echec du Drop sur le base de données %s"
#define MSG_DSORT_LOG_ERROR "Kindex: Erreur logique de tri distincte"
#define MSG_DUMMY_NO_COLS   "Les tables DUMMY ne peuvent pas avoir de colonne"
#define MSG_DUPLICAT_COUNT  "Count sur plus d'une colonne"
#define MSG_DUP_COL_NAME    "La colonne %s existe en double"
#define MSG_DUP_PROJNUM     "Non unique projnum %d pour la colonne %s"
#define MSG_DVAL_NOTIN_LIST "Valeur %s non trouvée dans la liste des valeurs distinctes de la colonne %s"
#define MSG_EMPTY_DOC       "Document vide"
#define MSG_EMPTY_FILE      "%s du fichier vide %s: "
#define MSG_ENDSTR_MISMATCH "Fins de chaîne et de noeud ne correspondent pas"
#define MSG_END_OF_DELETE   "%d ligne(s) enlevée(s) en %.2lf sec"
#define MSG_END_OF_INSERT   "%d ligne(s) insérée(s) en %.2lf sec"
#define MSG_END_OF_QUERY    "%d ligne(s) extraite(s) en %.2lf sec"
#define MSG_END_OF_UPDATE   "%d ligne(s) modifiée(s) en %.2lf sec"
#define MSG_EOF_AFTER_LINE  "Fin de fichier après la ligne %d"
#define MSG_EOF_INDEX_FILE  "EOF lisant le fichier index"
#define MSG_ERASED          " et effacée"
#define MSG_ERASE_FAILED    " (échec de l'effacement)"
#define MSG_ERROR           "Erreur"
#define MSG_ERROR_IN_LSK    "Erreur %d dans lseek64"
#define MSG_ERROR_IN_SFP    "Erreur %d dans SetFilePointer"
#define MSG_ERROR_NO_PARM   "Paramètre absent (valide seulement pour %.8s.1 et %.8s.5)"
#define MSG_ERROR_OPENING   "Erreur à l'ouverture de : "
#define MSG_ERR_NUM_GT_MAX  "Erreur: Numval (%d) plus grand que Maxnum (%d)"
#define MSG_ERR_READING_REC "Erreur lisant l'enregistrement %d de %s"
#define MSG_ERR_RET_RULE    "Retour erreur, règle=%u"
#define MSG_ERR_RET_TYPE    "Retour erreur, type=%d"
#define MSG_EVAL_EXPIRED    "Cette version d'évaluation est expirée"
#define MSG_EVAL_ONLY       "L'utilisation de cette Dll est pour évaluation seulement"
#define MSG_EXECUTING       "Exécution"
#define MSG_EXECUTION_ERROR "Erreur d'exécution"
#define MSG_EXEC_MODE_IS    "Le mode d'exécution est %s"
#define MSG_EXEC_MODE_RESET ". Mode remis à Execute"
#define MSG_EXEC_MODE_SET   "Mode d'exécution fixé à %s"
#define MSG_EXIT_EVAL_ERR   "Erreur pendant l'évaluation de Exit"
#define MSG_EXIT_FROM_LANG  "Fin du langage %s version %d.%d"
#define MSG_FAIL_ADD_NODE   "L'ajout du noeud %s dans la table a échoué"
#define MSG_FETCHING_DATA   "Recherche des données"
#define MSG_FETCHING_ROWS   "Recherche des lignes"
#define MSG_FETCH_NO_RES    "Fetch: Pas de Résultats"
#define MSG_FIELD_TOO_LONG  "Valeur trop longue pour le champs %d ligne %d"
#define MSG_FILELEN_ERROR   "Erreur dans %s pour %s"
#define MSG_FILE_CLOSE_ERR  "Erreur %d à la fermeture du fichier"
#define MSG_FILE_IS_EMPTY   "Le fichier %s est vide"
#define MSG_FILE_MAP_ERR    "Erreur de File mapping"
#define MSG_FILE_MAP_ERROR  "CreateFileMapping %s erreur rc=%d"
#define MSG_FILE_NOT_FOUND  "Fichier %s introuvable"
#define MSG_FILE_OPEN_YET   "Fichier %s déjà ouvert"
#define MSG_FILE_UNFOUND    "Fichier %s non trouvé"
#define MSG_FILGRP_NO_TABLE "Table %d manquante pour groupe filtre"
#define MSG_FILTER_ATTACH   "Filtre passé à Attach"
#define MSG_FILTER_NO_TABLE "Filtre: première table manquante"
#define MSG_FIND_BAD_TYPE   "Recherche dans un tableau: type non conforme %s %s"
#define MSG_FIX_OVFLW_ADD   "Dépassement de capacité en addition"
#define MSG_FIX_OVFLW_TIMES "Dépassement de capacité en mutiplication"
#define MSG_FIX_UNFLW_ADD   "Sous dépassement de capacité en addition"
#define MSG_FIX_UNFLW_TIMES "Sous dépassement de capacité en multiplication"
#define MSG_FLD_TOO_LNG_FOR "Champs %d trop long pour %s ligne %d de %s"
#define MSG_FLTST_NO_CORREL "FilTest ne devrait être appelé que pour les sous-requêtes corrélées"
#define MSG_FLT_BAD_RESULT  "Virgule flottante: résultat inexacte"
#define MSG_FLT_DENORMAL_OP "Opérande virgule flottante non normalisé"
#define MSG_FLT_INVALID_OP  "Opération virgule flottante invalide"
#define MSG_FLT_OVERFLOW    "Dépassement de capacité virgule flottante"
#define MSG_FLT_STACK_CHECK "Virgule flottante: Erreur de la pile"
#define MSG_FLT_UNDERFLOW   "Sous-dépassement de capacité virgule flottante"
#define MSG_FLT_ZERO_DIVIDE "Virgule flottante: division par zéro"
#define MSG_FMT_WRITE_NIY   "L'écriture des fichiers %s n'est pas encore implémentée"
#define MSG_FNC_NOTIN_SLIST "Fonction de tri absente de la liste de sélection"
#define MSG_FORMAT_ERROR    "Erreur de formattage"
#define MSG_FOXPRO_FILE     "Fichier FoxPro: "
#define MSG_FPUTS_ERROR     "Erreur dans fputs: %s"
#define MSG_FSBPARP_NULL    "PUTFON: fsbparp est nul"
#define MSG_FSEEK_ERROR     "Erreur dans fseek: %s"
#define MSG_FSETPOS_ERROR   "Erreur dans fseek pour i=%d"
#define MSG_FTELL_ERROR     "Erreur dans ftell enregistrement=%d: %s"
#define MSG_FUNCTION_ERROR  "Erreur dans %s: %d"
#define MSG_FUNC_ERRNO      "Erreur %d dans %s"
#define MSG_FUNC_ERROR      "Erreur dans %s"
#define MSG_FUNC_ERR_S      "Erreur dans %s: %s"
#define MSG_FUNC_REF_DEL    "Référence à une fonction définie (règle %d) qui a été supprimée"
#define MSG_FWRITE_ERROR    "Erreur dans fwrite: %s"
#define MSG_GETCWD_ERR_NO   "?getcwd %s errno=%d"
#define MSG_GETFILESIZE_ERR "Erreur %d dans GetFileSize"
#define MSG_GET_DIST_VALS   "Récupération des valeurs distinctes de "
#define MSG_GET_ERROR       "Erreur dans %s (colonne %d)"
#define MSG_GET_FUNC_ERR    "Erreur en recherche de la fonction %s: %s"
#define MSG_GET_NAME_ERR    "Erreur en retrouvant le nom d'une table SYS"
#define MSG_GLOBAL_ERROR    "Erreur d'allocation de Global (taille=%d)\n"
#define MSG_GRAM_ALLOC_ERR  "Erreur d'allocation dans Grammar Up"
#define MSG_GRAM_MISMATCH   "Avertissement: version de GRAMMAR perimée (sauvé sous GRAMMAR v%u)"
#define MSG_GRAM_SUBSET_ERR "Erreur d'initialisation du dictionnaire de la grammaire"
#define MSG_GRBY_TAB_NOTIMP "Group by avec tables jointes non implémenté"
#define MSG_GROUPBY_NOT_ALL "Group By doit inclure toutes les sélections non-fonctionnelles"
#define MSG_GROUP_ON_FUNC   "Group by invalide sur colonne fonctionnelle"
#define MSG_GRP_COL_MISM    "Disparité colonne des groupes"
#define MSG_GRP_LIST_MISMAT "Le groupement ne couvre pas la liste de sélection"
#define MSG_GUARD_PAGE      "Violation de page de garde"
#define MSG_GZOPEN_ERROR    "gzopen %s: erreur %d sur %s"
#define MSG_GZPUTS_ERROR    "Erreur dans gzputs: %s"
#define MSG_HANDLE_IS_NULL  "%s est NULL: erreur code: %d"
#define MSG_HARRY_COMP_NIY  "Compute non implémenté pour les chaînes codées"
#define MSG_HAVING_FILTER   "Traitement du Filtre Having"
#define MSG_HBUF_TOO_SMALL  "Buffer(%d) trop petit pour entête(%d)"
#define MSG_HEAD_OPEN_ERROR "Erreur à l'ouverture du fichier header"
#define MSG_HEAD_READ_ERROR "Erreur en lecture du fichier header %s"
#define MSG_HEAD_WRITE_ERR  "Erreur en écriture du fichier header"
#define MSG_HI_OFFSET_ERR   "Offset supérieur non nul"
#define MSG_HUGE_DEFAULT    "Huge est %d par défault"
#define MSG_HUGE_WARNING_1  "Mémoire Huge non compatible 16-bit pour %d\n"
#define MSG_HUGE_WARNING_2  "Résultats imprévisibles possibles\n"
#define MSG_IDLE            "Au repos"
#define MSG_ILLEGAL_INSTR   "Instruction illégale"
#define MSG_ILL_FILTER_CONV "Conversion implicite illégale dans un filtre"
#define MSG_INDEX_CREATED   "Index %s créé sur %s"
#define MSG_INDEX_DEF_ERR   "Erreur sauvegardant l'index définition pour %s"
#define MSG_INDEX_DROPPED   "Index %s supprimé de %s"
#define MSG_INDEX_INIT_ERR  "Echec de l'initialisation de l'index %s"
#define MSG_INDEX_NOT_DEF   "Index %s non défini"
#define MSG_INDEX_NOT_UNIQ  "L'index n'est pas Unique"
#define MSG_INDEX_ONE_SAVE  "Les index sont sauvegardés dans un fichier unique"
#define MSG_INDEX_SEP_SAVE  "Les index sont sauvegardés dans des fichiers séparés"
#define MSG_INDEX_YET_ON    "L'index %s existe déjà sur %s"
#define MSG_INDX_ALL_DROP   "Tous les index de %s supprimés"
#define MSG_INDX_COL_NOTIN  "La colonne index %s n'existe pas dans la table %s"
#define MSG_INDX_EXIST_YET  "L'entrée index existe déjà"
#define MSG_INIT_ERROR      "Erreur à l'initialisation de %s"
#define MSG_INIT_FAILED     "L'initialisation de %s a échoué"
#define MSG_INPUT           "Entrée: "
#define MSG_INPUT_KEYBD_YET "L'entrée est déjà au clavier"
#define MSG_INSERTING       "Insertion: "
#define MSG_INSERT_ERROR    "Insert erreur: usage multiple du fichier %s"
#define MSG_INSERT_MISMATCH "Les listes colonne et valeur ne correspondent pas"
#define MSG_INTERNAL        "interne"
#define MSG_INT_COL_ERROR   "Erreur interne sur la colonne index %s"
#define MSG_INT_OVERFLOW    "Dépassement de capacité sur entier"
#define MSG_INT_ZERO_DIVIDE "Division entière par zéro"
#define MSG_INVALID_BIP     "Bip invalide .%d"
#define MSG_INVALID_DISP    "Disposition invalide"
#define MSG_INVALID_FTYPE   "SBV: Ftype %d invalide"
#define MSG_INVALID_HANDLE  "Poignée invalide"
#define MSG_INVALID_OPER    "Opérateur invalide %d pour %s"
#define MSG_INVALID_OPTION  "Option invalide %s"
#define MSG_INV_COLUMN_TYPE "Type %d Invalide pour la colonne %s"
#define MSG_INV_COL_DATATYP "Type de données %d invalide pour la colonne %d"
#define MSG_INV_COL_NUM     "Colonne invalide %d"
#define MSG_INV_COL_TYPE    "Type de colonne %s invalide"
#define MSG_INV_CONC_BIP    "Bip invalide (seuls valides: %.8s.0  .1 and .5)"
#define MSG_INV_DATA_PATH   "Chemin vers les données invalide"
#define MSG_INV_DEF_READ    "Lecture différée invalide rc=%d"
#define MSG_INV_DIRCOL_OFST "Offset invalide pour une colonne DIR"
#define MSG_INV_DOMAIN_TYPE "Type invalide %d"
#define MSG_INV_FILTER      "Filtre résiduel dans %s"
#define MSG_INV_FNC_BUFTYPE "FNC: Type %d de l'argument invalide pour %s"
#define MSG_INV_INFO_TYPE   "Type d'info catalog invalide %d"
#define MSG_INV_INIPATH     "Inipath invalide "
#define MSG_INV_MAP_POS     "Position mémoire invalide"
#define MSG_INV_OPERATOR    "opérateur invalide %d\n"
#define MSG_INV_PARAMETER   "Paramètre invalide %s"
#define MSG_INV_PARM_TYPE   "Type de paramètre invalide"
#define MSG_INV_QUALIFIER   "Qalificateur '%s' invalide"
#define MSG_INV_QUERY_TYPE  "Type de requête %d invalide"
#define MSG_INV_RAND_ACC    "L'accès aléatoire d'une table non optimisée est impossible"
#define MSG_INV_REC_POS     "Position d'enregistrement invalide"
#define MSG_INV_RESULT_TYPE "Type de résultat invalide %s"
#define MSG_INV_SET_SUBTYPE "Type de formattage %d invalide"
#define MSG_INV_SPECIAL_CMD "%s: Commande spéciale invalide"
#define MSG_INV_SUBTYPE     "Sous type invalide %s"
#define MSG_INV_TOK_DOMAIN  "Le domaine %s n'existe pas"
#define MSG_INV_TOPSEM_CMD  "Commande TopSem invalide %c"
#define MSG_INV_TRANSF_USE  "Usage invalide en règle transformationnelle"
#define MSG_INV_TYPE_SPEC   "Spécification de type invalide (%.8s.%d)"
#define MSG_INV_UPDT_TABLE  "Table %s invalide pour Update"
#define MSG_INV_VALUE_LIST  "Liste de valeurs invalide pour Insert"
#define MSG_INV_WHERE_JOIN  "Clause Where invalide dans une requête de jointure"
#define MSG_INV_WORK_PATH   "Chemin de travail invalide"
#define MSG_IN_ARGTYPE_MISM "Arguments de types incompatibles pour une expression IN"
#define MSG_IN_USE          " et en activité"
#define MSG_IN_WITHOUT_SUB  "IN ou EXISTS sans tableau ou subquery"
#define MSG_IS_NOT_CONN     "%s n'est pas une connexion définie"
#define MSG_JCT_MISS_COLS   "Colonnes manquantes pour une table JCT"
#define MSG_JCT_MISS_TABLE  "Table jointe manquante pour JCT"
#define MSG_JCT_NO_FILTER   "Filtrage impossible des tables virtuelles JCT"
#define MSG_JCT_NO_KEY      "Erreur logique JCT: clé manquante"
#define MSG_JOIN_KEY_NO_COL "La clé de jointure n'est pas une colonne"
#define MSG_KEY_ALLOC_ERR   "Erreur d'allocation d'un bloc offset clé"
#define MSG_KEY_ALLOC_ERROR "Erreur d'allocation mémoire, Klen=%d n=%d"
#define MSG_LANGUAGE_QUIT   "%s libéré"
#define MSG_LANG_ACTIVE     "Langage %s actif"
#define MSG_LANG_ALLOC_FAIL "PlugInitLang: Erreur d'allocation du bloc Lang"
#define MSG_LANG_ALREADY_UP "Langage déjà en édition"
#define MSG_LANG_BAD_SAVE   "Langage %s peut-être incorrectement sauvegardé"
#define MSG_LANG_NOT_FREED  "Langage %s non libérable (pas dans la chaîne principale)"
#define MSG_LANG_SAVED      "Langage %s sauvegardé"
#define MSG_LANG_WR_LEN_ERR "Erreur de longueur à l'écriture du bloc Lang"
#define MSG_LDF_ALLOC_ERROR "Erreur d'allocation d'un LdfBlock"
#define MSG_LDF_RN_MISMATCH "LDF: décalage des numéros de règle"
#define MSG_LDF_WLEN_ERROR  "Erreur de longueur en écrivant LdfData"
#define MSG_LDF_W_LEN_ERROR "Erreur de longueur pour LdfData en écriture"
#define MSG_LIC_NO_MYSQL    "Votre licence actuelle ne permet pas l'utilisation du type MYSQL"
#define MSG_LINEAR_ERROR    "Erreur de linéarisation"
#define MSG_LINE_LENGTH     "Largeur d'impression fixée à %d"
#define MSG_LINE_MAXLIN     "Nombre de lignes de travail plafonné à %d"
#define MSG_LINE_MAXRES     "Nombre de lignes de résultat plafonné à %d"
#define MSG_LINE_MAXTMP     "Nombre de lignes intermédiaires plafonné à %d"
#define MSG_LINE_TOO_LONG   "La nouvelle ligne est trop longue"
#define MSG_LINJOINDB_ERROR "Erreur système: appel incorrecte à LinJoinDB"
#define MSG_LIST            "--Liste--"
#define MSG_LNG_NOT_IN_LIST "Le langage %s n'est pas dans la liste"
#define MSG_LOADING_DB      "Chargement description de la BD"
#define MSG_LOADING_FAILED  "Le chargement de %s a échoué"
#define MSG_LOAD_CDLL_ERROR "Erreur au chargement de ConnDll: rc=%d"
#define MSG_LOCSTRG_TOO_BIG "LOCSTRG: n trop grand ? (%d)\n"
#define MSG_LOGICAL_ERROR   "%s: Erreur logique"
#define MSG_LRECL_TOO_SMALL "Lrecl trop petit (longueur en-tête = %d)"
#define MSG_MAC_NO_DELETE   "Pas de suppression de lignes pour les tables MAC"
#define MSG_MAC_NO_INDEX    "Pas d'accès direct aux tables MAC"
#define MSG_MAC_READ_ONLY   "Les tables MAC sont en lecture seulement"
#define MSG_MAC_WIN_ONLY    "Les tables MAC sont seulement sous Windows"
#define MSG_MAKE_EMPTY_FILE "Génération du fichier vide %s: %s"
#define MSG_MAKING          "Génération"
#define MSG_MAKING_DISTINCT "Regroupement des valeures distinctes"
#define MSG_MALLOC_ERROR    "Allocation mémoire impossible par %s"
#define MSG_MALLOC_NULL     "malloc retourne Null"
#define MSG_MAP_NO_MORE     "Le type %s n'est plus supporté"
#define MSG_MAP_OBJ_ERR     "Erreur %d à la fermeture du map objet"
#define MSG_MAP_VEC_ONLY    "MAP Insert permis seulement pour les tables VEC Estimate"
#define MSG_MAP_VIEW_ERROR  "MapViewOfFile %s erreur rc=%d"
#define MSG_MAXSIZE_ERROR   "Maxsize incalculable sur table ouverte"
#define MSG_MAXTMP_TRUNCATE "Résultats intermédiaires tronqués par maxtmp=%d"
#define MSG_MAX_BITMAP      "Taille maxi des bitmaps d'optimisation  fixée à %d"
#define MSG_MEMSIZE_TOO_BIG "Erreur: memsize (%d) trop grand pour Length (%d)"
#define MSG_MEM_ALLOC_ERR   "Erreur d'allocation mémoire, taille %s = %d"
#define MSG_MEM_ALLOC_ERROR "Erreur d'allocation mémoire"
#define MSG_MEM_ALLOC_YET   "Mémoire déjà allouée"
#define MSG_METAFILE_NOTFND "Fichier Meta introuvable"
#define MSG_MISPLACED_QUOTE "Appostrophe mal placée ligne %d"
#define MSG_MISSING         "Manquant: Value=%p Argval=%p Builtin=%d"
#define MSG_MISSING_ARG     "Argument manquant pour l'opérateur %d"
#define MSG_MISSING_COL_DEF "Définition des colonnes manquante"
#define MSG_MISSING_CONNECT "Connection #1 manquante"
#define MSG_MISSING_EOL     "Fin de ligne manquante dans %s"
#define MSG_MISSING_FIELD   "Champs %d manquant dans %s ligne %d"
#define MSG_MISSING_FNAME   "Nom du fichier manquant"
#define MSG_MISSING_NODE    "Noeud %s manquant dans %s"
#define MSG_MISSING_POS     "POS code manquant"
#define MSG_MISSING_ROWNODE "Impossible de trouver le noeud de la ligne %d"
#define MSG_MISSING_SERV_DB "Indication serveur et/ou base de données manquante"
#define MSG_MISS_LEAD_COL   "Colonne majeure %s manquante"
#define MSG_MISS_NAME_LRECL "Nom du fichier et/ou LRECL manquant"
#define MSG_MISS_TABLE_LIST "Liste des tables manquante"
#define MSG_MISS_VCT_ELMT   "Taille de bloc vectoriel manquante (Elements)"
#define MSG_MIS_TAG_LIST    "Liste des balises colonne manquante"
#define MSG_MKEMPTY_NIY     "MakeEmptyFile: pas encore implementé pour Huge et Unix"
#define MSG_MOVE_INV_TYPE   "MOVPARM: paramètre de type invalide %d"
#define MSG_MULT_DISTINCT   "Distinct utilisé plus d'une fois"
#define MSG_MULT_KEY_ERROR  "Erreur sur clé multiple k=%d n=%d"
#define MSG_MUL_MAKECOL_ERR "Erreur logique dans TABMUL::MakeCol"
#define MSG_MYSQL_CNC_OFF   "La connexion à MySQL est fermée"
#define MSG_MYSQL_CNC_ON    "La connexion à MySQL est établie"
#define MSG_MYSQL_NOT_SUP   "Pas de support de MySQL dans cette version"
#define MSG_MY_CNC_ALREADY  "La connexion à MySQL est déjà active"
#define MSG_NAME_CONV_ERR   "Erreur de convertion du nom de noeud"
#define MSG_NAME_IS_USED    "Le nom %s est déjà utilisé"
#define MSG_NCOL_GT_MAXCOL  "Trop de colonnes (%d > %d max)"
#define MSG_NEW_CHAR_NULL   "new char(%d) retourne Null"
#define MSG_NEW_DOC_FAILED  "Impossible de créer le nouveau document"
#define MSG_NEW_RETURN_NULL "NULL renvoyé par New dans PlugEvalLike"
#define MSG_NEW_TABLE_ERR   "La nouvelle table %s ne peut pas être chargée"
#define MSG_NEXT_FILE_ERROR "Erreur en recherche du fichier suivant. rc=%s"
#define MSG_NODEF_FROM_VIEW "Pas de définition de table depuis une view"
#define MSG_NODE_FOR_CHAR   "Noeud %s trouvé au lieu d'un caractère"
#define MSG_NODE_SUBSET_ERR "Erreur d'initialisation de la zone Noeud %d"
#define MSG_NONCONT_EXCEPT  "Exception non-continuable"
#define MSG_NON_DUP_HAVING  "Clause Having dans une requête non fonctionelle"
#define MSG_NON_EVAL_SEM    "Sem non évaluée: p_no=%d"
#define MSG_NOP_ZLIB_INDEX  "L'indexage d'une table zlib non optimisée est impossible"
#define MSG_NOT_A_DBF_FILE  "Le fichier n'a pas le format dBASE dbf "
#define MSG_NOT_ENOUGH_COLS "Pas assez de colonnes dans %s"
#define MSG_NOT_ENOUGH_MEM  "Mémoire insuffisante pour cette opération"
#define MSG_NOT_FIXED_LEN   "Fichier %s non fixe, len=%d lrecl=%d"
#define MSG_NOT_IMPLEMENTED "Non implementé: %.8s"
#define MSG_NOT_IMPL_JOIN   "Pas implémenté pour les jointures"
#define MSG_NOT_IMPL_SET    "Pas implémenté pour les opérateurs d'ensembles"
#define MSG_NOT_IMPL_YET    "Pas encore implementé"
#define MSG_NOT_LINEARIZED  "Arborescence des tables non linéarisée"
#define MSG_NOT_MODIFIABLE  " (non modifiable)"
#define MSG_NO_0DH_HEAD     "0DH manquant en fin d'en-tête (dbc=%d)"
#define MSG_NO_ACTIVE_APPL  "Pas d'application active"
#define MSG_NO_ACTIVE_DB    "Pas de base de données active"
#define MSG_NO_ACTIVE_UDIC  "Pas de dictionaire utilisateur actif"
#define MSG_NO_AGGR_FUNC    "Fonction aggrégée %d illégale à cet endroit"
#define MSG_NO_AREA_FILE    "Fichier Area introuvable"
#define MSG_NO_AVAIL_RESULT "Pas de résultat disponible"
#define MSG_NO_BIG_DELETE   "Délétion Partielle non implémentée pour les fichiers HUGE"
#define MSG_NO_CHAR_FROM    "Conversion de type %d en caractères impossible"
#define MSG_NO_CLUSTER_COL  "Pas de colonne optimisable"
#define MSG_NO_COL_ADDING   "Ajouter des colonnes dans une définition existante est impossible"
#define MSG_NO_COL_DEF_AS   "La définitions des colonnes est incompatible avec AS Select"
#define MSG_NO_COL_FOUND    "La section colonne %s est vide"
#define MSG_NO_COL_IN_TABLE "La colonne %d n'est pas dans la table %s"
#define MSG_NO_COL_SECTION  "Section colonne manquante pour la table %s"
#define MSG_NO_CONNECT_ADDR "Adresse de connection non spécifiée"
#define MSG_NO_CONST_FILTER "Filtres constants non implementés"
#define MSG_NO_CURLY_BRKT   "Pas d'accolade de fermeture"
#define MSG_NO_DATABASE     "Base de données %s introuvable"
#define MSG_NO_DATE_FMT     "Pas de format date pour le valblock de type %d"
#define MSG_NO_DBF_INSERT   "Insert pas encore implémenté pour les fichier DBF"
#define MSG_NO_DEF_FNCCOL   "Colonne fonction par défaut introuvable"
#define MSG_NO_DEF_PIVOTCOL "Colonne pivot par défaut introuvable"
#define MSG_NO_DIR_INDX_RD  "Pas d'accès directe des tables %s"
#define MSG_NO_DMY_DIR_ACC  "Pas d'accès direct aux tables virtuelles DUMMY"
#define MSG_NO_DOM_DELETE   "Délétion Partielle non implémentée pour les domaines"
#define MSG_NO_DOM_MATCH    "Chaîne %.8s... non touvée dans le domaine %s"
#define MSG_NO_EDITED_LANG  "Coparm: Pas de langage en édition"
#define MSG_NO_EXP_LINK     "Liaison par expression invalide pour une table JCT"
#define MSG_NO_EXT_FILTER   "Le filtrage ne peut se référer à une autre table"
#define MSG_NO_EXT_UPDATE   "Pas de mise à jour en référence à une autre table"
#define MSG_NO_FEAT_SUPPORT "%s non supporté dans cette version"
#define MSG_NO_FILE_LIST    "La table %s n'a pas de liste de fichiers"
#define MSG_NO_FLD_FORMAT   "Format absent pour le champs %d de %s"
#define MSG_NO_FORMAT_COL   "Type COLUMN informattable"
#define MSG_NO_FORMAT_TYPE  "Le format ne peut pas être défini à partir du type %d"
#define MSG_NO_FULL_JOIN    "Jointures autorisées seulement à égalité sur clé(s)"
#define MSG_NO_FUL_OUT_JOIN "Jointures externes complètes non supportées"
#define MSG_NO_FUNC_ORDER   "Tri non supporté sur élément fonctionnel"
#define MSG_NO_HEAD_JOIN    "Jointure sur une table non en tête"
#define MSG_NO_HQL_CONV     "Conversion en HQL non disponible"
#define MSG_NO_INDEX        "La table %s n'a pas d'index"
#define MSG_NO_INDEX_GBX    "Pas ou mauvais index pour SQLGBX"
#define MSG_NO_INDEX_IN     "Pas d'index dans %s"
#define MSG_NO_INDEX_READ   "Pas d'accès directe des tables multiples"
#define MSG_NO_INIT_LANG    "Pas de langage initial"
#define MSG_NO_JOIN_TO_EXP  "Jointure vers une expression impossible"
#define MSG_NO_JOIN_UPDEL   "Pas de jointure avec Update/Delete"
#define MSG_NO_KEY_COL      "Pas de colonne clé trouvée"
#define MSG_NO_KEY_UPDATE   "Le nom des clés ne peut pas être modifié"
#define MSG_NO_LANGUAGE     "Pas de langage opérationnel\n"
#define MSG_NO_LANG_TO_QUIT "Pas de langage à quitter"
#define MSG_NO_LISTVAL_HERE "LSTBLK: Liste de valeurs utilisée hors contexte"
#define MSG_NO_MAP_INSERT   "MAP incompatible avec Insert"
#define MSG_NO_MATCHING_COL "Pas de colonne correspondant à %s dans %s"
#define MSG_NO_MATCH_COL    "Colonne correspondante introuvable"
#define MSG_NO_MEMORY       "Mémoire pleine"
#define MSG_NO_MEM_CORR_SUB "Subquery corrélée en mémoire non encore implémentée"
#define MSG_NO_MODE_PADDED  "Mode non supporté pour les fichiers 'padded'"
#define MSG_NO_MORE_COL     "La colonne %s n'est plus dans la table pivot"
#define MSG_NO_MORE_LANG    "Plus de langage, exit de %s\n"
#define MSG_NO_MORE_VAR     "Les fichiers VAR ne sont plus supportés"
#define MSG_NO_MULCOL_JOIN  "Jointure vers un index multi-colonne pas encore possible"
#define MSG_NO_MULT_HAVING  "Clauses Having multiples non implémentées"
#define MSG_NO_MUL_DIR_ACC  "Accès direct des tables multiples pas encore implémenté"
#define MSG_NO_MUL_VCT      "Les tables VCT ne peuvent pas être multiples"
#define MSG_NO_MYSQL_CONN   "Aucune connexion MySQL ouverte"
#define MSG_NO_MYSQL_DELETE "Pas de Delete pour les tables MySQL"
#define MSG_NO_NBCOL        "Pas de NBcol"
#define MSG_NO_NBLIN        "Pas de NBlin, MaxSize ou Continued"
#define MSG_NO_NBLIN_CONT   "Fetch: Pas de NBlin ou Continued"
#define MSG_NO_NULL_CONST   "Les constantes <null> ne sont pas prises en charge"
#define MSG_NO_ODBC_COL     "Colonnes ODBC automatiques non supportées par cette version"
#define MSG_NO_ODBC_DELETE  "Delete ne devrait pas être appelé pour les tables ODBC"
#define MSG_NO_ODBC_DIRECT  "Accès directe des tables ODBC non encore implémenté"
#define MSG_NO_ODBC_MUL     "Multiple(2) non supporté pour les tables ODBC"
#define MSG_NO_ODBC_SPECOL  "Pas de colonne spéciale ODBC"
#define MSG_NO_OPT_COLUMN   "Pas optimisable ou pas de colonne optimisées"
#define MSG_NO_OP_MODIF     "Les modificateurs ne s'appliquent pas à %s"
#define MSG_NO_PARAMETER    "Pas de paramètre"
#define MSG_NO_PART_DEL     "Delete partiel des fichier %s impossible"
#define MSG_NO_PART_MAP     "Mapping partiel non implémenté pour cet OS"
#define MSG_NO_PAR_BLK_INS  "Insertion de bloc partiel impossible"
#define MSG_NO_PIV_DIR_ACC  "Pas d'accès directe aux tables PIVOT"
#define MSG_NO_POS_ADDED    "Pos_code non ajouté"
#define MSG_NO_PROMPTING    "Relance impossible pour les tables distribuées"
#define MSG_NO_QRY_DELETE   "Delete n'est pas utilisable pour les views QRY"
#define MSG_NO_QUERY_ARRAY  "Tableaux avec QUERY non encore implémentés"
#define MSG_NO_RCUR_DSK_YET "Usage recursif de DISK non encore implementé"
#define MSG_NO_READ_32      "Lecture de 32 octets impossible"
#define MSG_NO_RECOV_SPACE  "Espace non recouvrable dans le fichier index"
#define MSG_NO_REF_DELETE   "Pas de suppression en référence à une autre table"
#define MSG_NO_REF_UPDATE   "Pas de mise à jour en référence à une autre table"
#define MSG_NO_REMOTE_FNC   "Certaines fonctions ne peuvent pas être exécutées à distance"
#define MSG_NO_ROWID_FOR_AM "Accès direct impossible de ROWID pour les tables de type %s"
#define MSG_NO_ROW_NODE     "Le nom du Rownode n'est pas défini"
#define MSG_NO_SECTION_NAME "Nom de section manquant"
#define MSG_NO_SEC_UPDATE   "Les noms de section ne peuvent pas être modifiés"
#define MSG_NO_SELECTED_DB  "Aucune base de données sélectée"
#define MSG_NO_SELF_PIVOT   "Une table ne peut se pivoter elle-même !"
#define MSG_NO_SERVER_FOUND "Serveur introuvable"
#define MSG_NO_SETPOS_YET   "SetPos pas encore implémenté pour les fichier %s"
#define MSG_NO_SFEXIT_UNIX  "Fonction %s non disponible sur Unix"
#define MSG_NO_SOURCE       " (pas de source)"
#define MSG_NO_SPEC_COL     "Pas de colonne spéciales MYSQL"
#define MSG_NO_SQL_DELETE   "Delete n'est pas utilisable actuellement pour les views SQL"
#define MSG_NO_SUB_VAL      "Pas de sous-value d'un tableau de type %d"
#define MSG_NO_SUCH_INDEX   "La table %s n'a pas l'index %s"
#define MSG_NO_SUCH_SERVER  "Serveur %s introuvable"
#define MSG_NO_SUCH_TABLE   "Table %s pas dans la base de données"
#define MSG_NO_TABCOL_DATA  "Pas de données pour la table %s colonne %s"
#define MSG_NO_TABLE_COL    "Aucune colonne trouvée pour %s"
#define MSG_NO_TABLE_DEL    "Delete non autorisé pour les tables %s   "
#define MSG_NO_TABLE_DESC   "Pas de bloc descriptif de table"
#define MSG_NO_TABLE_INDEX  "La table %s n'a pas d'index"
#define MSG_NO_TABLE_LIST   "Pas de liste de tables"
#define MSG_NO_TAB_DATA     "Pas de données pour la table %s"
#define MSG_NO_TERM_IN_TOK  "Les non-terminaux ne sont pas utilisables dans les règles de Token"
#define MSG_NO_TOKEN_DB     "DB introuvable pour la colonne TOKEN %s"
#define MSG_NO_UNIX_CATINFO "Pas d'info catalogue sous Unix"
#define MSG_NO_UPDEL_JOIN   "Pas de jointure de tables ODBC pour Update/Delete"
#define MSG_NO_VCT_DELETE   "Délétion Partielle non implémentée pour les fichiers VCT"
#define MSG_NO_VIEW_COLDEF  "Colonne définition impossible pour les views"
#define MSG_NO_VIEW_SORT    "La View fonctionnelle %s ne peut pas être triée ou jointe"
#define MSG_NO_ZIP_DELETE   "Delete sur fichier Zip non encore implementé"
#define MSG_NO_ZIP_DIR_ACC  "Accès directe des tables ZDOS non encore implementé"
#define MSG_NULL_COL_VALUE  "La colonne n'a pas de valeur"
#define MSG_NULL_ENTRY      "InitLang, entrée nulle %d %s"
#define MSG_NULL_QUERY      "Requête vide"
#define MSG_NUMVAL_NOMATCH  "Disparité de Numval pour %s"
#define MSG_N_FULL_PARSES   "%d significations"
#define MSG_ODBC_READ_ONLY  "ODBC est actuellement en lecture seulement"
#define MSG_OFFSET_NOT_SUPP "Offset non supporté pour ce type de sous requête"
#define MSG_ONE_LANG_YET    "Un langage est déjà en édition"
#define MSG_ONE_PARAM_ONLY  "Un seul paramètre autorisé"
#define MSG_ONLY_LOG10_IMPL "Seul Log10 est implementé"
#define MSG_ON_LANGUAGE     "Langage %.8s version %d niveau %d éditable"
#define MSG_OPENING         "Ouverture"
#define MSG_OPENING_QUERY   "Ouverture de la requête"
#define MSG_OPEN_EMPTY_FILE "Ouverture du fichier vide %s: %s"
#define MSG_OPEN_ERROR      "Erreur d'ouverture %d en mode %d sur %s: "
#define MSG_OPEN_ERROR_IS   "Erreur à l'ouverture de %s: %s"
#define MSG_OPEN_ERROR_ON   "Erreur d'ouverture sur %s"
#define MSG_OPEN_MODE_ERROR "Erreur d'ouverture(%s) %d sur %s"
#define MSG_OPEN_SORT_ERROR "Erreur logique de tri dans QUERY Open"
#define MSG_OPEN_STRERROR   "Erreur à l'ouverture: %s"
#define MSG_OPEN_W_ERROR    "Erreur à l'ouverture de %s en écriture"
#define MSG_OPTBLK_RD_ERR   "Erreur à la lecture d'un bloc optimisation: %s"
#define MSG_OPTBLK_WR_ERR   "Erreur à l'écriture d'un bloc optimisation: %s"
#define MSG_OPTIMIZING      "Optimisation de "
#define MSG_OPT_BMAP_RD_ERR "Erreur en lecture des bitmaps d'optimisation: %s"
#define MSG_OPT_BMAP_WR_ERR "Erreur en écriture des bitmaps d'optimisation: %s"
#define MSG_OPT_CANCELLED   "Optimisation interrompue par l'utilisateur"
#define MSG_OPT_DVAL_RD_ERR "Erreur en lecture des valeurs distinctes: %s"
#define MSG_OPT_DVAL_WR_ERR "Erreur en écriture des valeurs distinctes: %s"
#define MSG_OPT_HEAD_RD_ERR "Erreur en lecture de l'entête du fichier opt: %s"
#define MSG_OPT_HEAD_WR_ERR "Erreur en écriture de l'entête du fichier opt: %s"
#define MSG_OPT_INIT        "Optimisation initialisée"
#define MSG_OPT_LOGIC_ERR   "Erreur logique dans SetBitmap, i=%d"
#define MSG_OPT_MAX_RD_ERR  "Erreur en lecture des valeurs maxi: %s"
#define MSG_OPT_MAX_WR_ERR  "Erreur en écriture des valeurs maxi: %s"
#define MSG_OPT_MIN_RD_ERR  "Erreur en lecture des valeurs mini: %s"
#define MSG_OPT_MIN_WR_ERR  "Erreur en écriture des valeurs mini: %s"
#define MSG_OPT_NOT_MATCH   "Le fichier opt %s n'est pas à jour"
#define MSG_OP_RES_TOO_LONG "Résultat trop long pour l'opérateur=%d"
#define MSG_ORDER_OUT_RANGE "Tri: Order %d hors limite"
#define MSG_ORDER_TWICE     "Un même élément est trié deux fois"
#define MSG_PAGE_ERROR      "Erreur de pagination"
#define MSG_PARM_CNT_MISS   "Disparité du nombre de Paramètres"
#define MSG_PARSE_NULL_SEM  "Sémantique nulle"
#define MSG_PARSING_QUERY   "Analyse de la requête"
#define MSG_PIX_ERROR       "Pix %s erreur règle no=%u\n"
#define MSG_PIX_TEST_ERROR  "Règle=%u: pix-TEST pas dans le premier noeud\n"
#define MSG_PLG_READ_ONLY   "PLG est actuellement en lecture seulement"
#define MSG_PLM_NULL_SFP    "TABPLM ReadDB: Sfp est NULL"
#define MSG_PLUG_NOT_INIT   "Plug n'est pas initialisé\n"
#define MSG_PLUG_NOT_RUN    "Plug n'est pas en marche"
#define MSG_PNODE_RULE      "(Noeud %d règle %d) "
#define MSG_POS_TOO_LONG    "%s trop long (>%d)"
#define MSG_PREC_VBLP_NULL  "ARRAY SetPrecision: Vblp est NULL"
#define MSG_PRIV_INSTR      "Instruction privilégiée"
#define MSG_PROCADD_ERROR   "Erreur %d sur l'adresse de %s"
#define MSG_PROCESS_SUBQRY  "Sub-Query en cours de traitement"
#define MSG_PROC_WOULD_LOOP "Bouclage du traitement (maxres=%d maxlin=%d)"
#define MSG_PROGRESS_INFO   "Informations sur le traitement en cours"
#define MSG_PROMPT_CANCEL   "Relance annulée"
#define MSG_PROMPT_NIY      "Prompt non implémenté pour cette configuration"
#define MSG_PTR_NOT_FOUND   "Pointeur introuvable Num=%d ti1=%d"
#define MSG_PXDEF_IS_NULL   "Pxdef est NULL"
#define MSG_QRY_READ_ONLY   "Les views QRY sont en lecture seulement"
#define MSG_QUERY_CANCELLED "Requête interrompue par l'utilisateur"
#define MSG_QUERY_NOT_EXEC  "Requête non exécutée"
#define MSG_QUERY_SAVED     "Requête %s sauvegardée"
#define MSG_QUOTE_IN_QUOTE  "Appostrophe dans un champ entre appostrophe ligne %d"
#define MSG_RANGE_NIY       "Range pas encore implémenté pour %s"
#define MSG_RANGE_NO_JOIN   "Range non compatible avec les index de jointure"
#define MSG_RC_READING      "rc=%d en lecture de la table %s"
#define MSG_READB_BAD_INIT  "%s ReadDB appelé avec Init=0"
#define MSG_READCOL_ERROR   "SQLCOL: erreur dans ReadColumn"
#define MSG_READING         "Lecture"
#define MSG_READING_FROM    "Lecture de %s"
#define MSG_READING_RECORD  "Erreur en lecture de l'enregistrement %d de %s"
#define MSG_READY           "Prêt"
#define MSG_READ_ERROR      "Erreur en lecture sur %s: %s"
#define MSG_READ_ERROR_RC   "Erreur en lecture, rc=%d"
#define MSG_READ_MEM_ERROR  "Lecture mémoire %d: taille=%d"
#define MSG_READ_ONLY       "Cette table protégée en lecture seule ne peut être modifiée"
#define MSG_READ_SEEK_ERROR "Erreur de recherche en lecture: %s"
#define MSG_READ_SEG_ERROR  "Lecture segment %d: taille=%d"
#define MSG_RECEIVED        "Reçu %c\n"
#define MSG_RECORD_ERROR    "Erreur à la lecture de l'enregistrement %d de %s"
#define MSG_RECORD_NO_SEP   "Enregistrement sans séparateur"
#define MSG_REC_SKIPPED     " (%d lignes erronnées sautées par l'option MaxErr)"
#define MSG_REDUCE_INDEX    "Réduction de l'index"
#define MSG_REGISTER_ERR    "Enregistrement NS impossible, préfix='%s' et href='%s'"
#define MSG_REMOTE_CONN_ERR "La connection éloignée a échoué"
#define MSG_REMOVE_ERROR    "Erreur en supprimant %s: %s"
#define MSG_REMOVE_NOT_IMPL "Remove non implémenté pour TDB non Table"
#define MSG_RENAME_ERROR    "Erreur renommant %s en %s: %s"
#define MSG_RENUM_RULES     "Renumérotez les règles et réentrez ADD (règle sauvegardée dans la zone tampon)"
#define MSG_REORDER_INDEX   "Reclassement de l'index"
#define MSG_REQU_ARG_NUM    "La fonction %s doit avoir %d arguments"
#define MSG_RESET_TO        "%s remis à %d"
#define MSG_RES_NOT_UNIQUE  "Le résultat n'est pas unique"
#define MSG_RET_FROM_LANG   "Retour au language %s version %d.%d du language %s version %d.%d"
#define MSG_ROWID_NOT_IMPL  "RowNumber non implémenté pour les tables de type %s"
#define MSG_ROWS_SELECTED   "%d lignes sélectionnées en %.2lf sec"
#define MSG_ROWS_TRUNCATED  " (tronqué par MAXRES, LIMIT, FREQ ou AreaSize)"
#define MSG_ROW_ARGNB_ERR   "ROW: disparité du nombre d'arguments (%d,%d)"
#define MSG_RPC_SERVER_ERR  "Erreur logique dans TABMUL::MakeCol"
#define MSG_RSC_ALLOC_ERROR "Erreur d'allocation mémoire dans Rescol %s"
#define MSG_RULE_ENTERED    "Règle %d entrée"
#define MSG_RULE_SUBSET_ERR "Erreur d'initialisation de la zone Règles"
#define MSG_SAVING_INDEX    "Sauvegarde du fichier index"
#define MSG_SCAN_NOT_IMP    "Scan non implémenté"
#define MSG_SEC_KEY_FIRST   "Les sections et clés doivent être insérées en premier"
#define MSG_SEC_NAME_FIRST  "Le nom de section doit être en tête de liste en insertion"
#define MSG_SEC_NOT_FOUND   "Section %s absente de %s"
#define MSG_SEEK_ERROR      "Seek erreur dans CopyHeader"
#define MSG_SEMANTIC_TREE   "Arbre sémantique"
#define MSG_SEM_BAD_REF     "Sem @%d référence un argument de type non 0 ou 1"
#define MSG_SEM_UNKNOWN     "inconnue, rc=%d"
#define MSG_SEP_IN_FIELD    "Le champ %d contient le caractère séparateur"
#define MSG_SEQUENCE_ERROR  "HSTMT: Allocation hors séquence"
#define MSG_SETEOF_ERROR    "Erreur %d dans SetEndOfFile"
#define MSG_SETRECPOS_NIY   "SetRecpos non implémenté pour ce type de table"
#define MSG_SET_LOCALE      "Locale fixée à %s"
#define MSG_SET_NULL_DOM    "Valeur %d donnée à un domaine nul"
#define MSG_SET_OP_NOT_IMPL "Opérateurs ensemblistes non implementés"
#define MSG_SET_STR_TRUNC   "SetValue: Chaîne de caractères tronquée"
#define MSG_SEVERAL_TREES   "Jointure non spécifiée pour certaines tables"
#define MSG_SFP_ERROR       "Erreur sur SetFilePointer: %s"
#define MSG_SFUNC_NOT_IMPL  "Fonction scalaire %s non implémentée"
#define MSG_SHARED_LIB_ERR  "Erreur au chargement de la librairie partagée %s: %s"
#define MSG_SINGLE_STEP     "Pas à pas"
#define MSG_SLEEP           "J'ai dormi %d milliseconds"
#define MSG_SMART_SORTING   "Récupération des lignes triées (passage %d de %d)"
#define MSG_SMART_SORT_ERR  "Erreur logique 1 dans Smart Sort"
#define MSG_SORTING         "Tri en cours"
#define MSG_SORTING_INDEX   "Tri de l'index"
#define MSG_SORTING_VAL     "Tri de %d valeurs"
#define MSG_SORT_JOIN_INDEX "Tri de l'index de jointure"
#define MSG_SPCOL_READONLY  "La colonne spéciale %s est en lecture seulement"
#define MSG_SPEC_CMD_SEP    "Les commandes spéciales doivent être exécutées séparément"
#define MSG_SQL_BAD_TYPE    "RephraseSQL: type %d non supporté"
#define MSG_SQL_BLOCK_MISM  "CheckColumn: bloc SQL courant non correspondant"
#define MSG_SQL_CONF_ERROR  "Erreur SQL: SQL_CONFORMANCE"
#define MSG_SQL_READ_ONLY   "Les views SQL sont actuellement en lecture seulement"
#define MSG_SRCH_CLOSE_ERR  "Erreur à la fermeture de l'Handle de recherche"
#define MSG_SRC_TABLE_UNDEF "La table source n'est pas définie"
#define MSG_STACK_ERROR     "Erreur sur la pile, i=%d\n"
#define MSG_STACK_OVERFLOW  "Parser: Débordement de la pile\n"
#define MSG_STRG_NOT_FOUND  "Chaîne introuvable"
#define MSG_STRING_INV_LIST "Liste invalide pour SemString"
#define MSG_STRING_TOO_BIG  "Chaîne trop grande pour le domaine %s"
#define MSG_SUBALLOC_ERROR  "Pas assez de mémoire en zone %p pour allouer %d (utilisé=%d libre=%d)"
#define MSG_SUBAL_HUGE_ERR  "Pas assez de mémoire en zone huge %p pour allouer %d"
#define MSG_SUBARG_NOSEM    "Argument @ ou sous-phrase de niveau %d pointe sur un noeud sans Sem"
#define MSG_SUBARG_OUTRANGE "Argument @ ou sous-phrase de niveau %d hors limite"
#define MSG_SUBQRY_ONEITEM  "Une Sub-Query ne doit avoir qu'une sélection"
#define MSG_SUBSET_ERROR    "SubSet erreur dans LoadDB"
#define MSG_SUB_OPEN_YET    "Subquery déjà ouverte"
#define MSG_SUB_RES_TOO_LNG "Résultat trop long pour SUBSTR"
#define MSG_SYNTAX_ERROR    "Erreur de syntaxe"
#define MSG_SYSTEM_ERROR    "Erreur système %d"
#define MSG_S_ACCESS_DENIED "%s: accès non autorisé"
#define MSG_S_ERROR         "%s erreur"
#define MSG_S_ERROR_NUM     "%s: erreur=%d"
#define MSG_S_INTRUPT_ERROR "%s: erreur interruption"
#define MSG_S_INVALID_PARM  "%s: paramètre invalide"
#define MSG_S_INV_ADDRESS   "%s: adresse invalide"
#define MSG_S_UNKNOWN_ERROR "%s: erreur de code %u inconnu"
#define MSG_TABDIR_READONLY "Les tables DIR sont en lecture seulement"
#define MSG_TABLE_ALREADY   "La table %s existe déjà"
#define MSG_TABLE_ALTERED   "Table %s %s altérée"
#define MSG_TABLE_CREATED   "%s table %s créée"
#define MSG_TABLE_DROPPED   "Table %s supprimée"
#define MSG_TABLE_MULT_JOIN "Utilisation multiple de la table %s pour jointure"
#define MSG_TABLE_NOT_IN_DB "La table %s n'existe pas dans %s"
#define MSG_TABLE_NOT_OPT   "Table non optimisable"
#define MSG_TABLE_NO_INDEX  "La table %s n'est pas indexable"
#define MSG_TABLE_NO_OPT    "La table %s n'existe pas ou de type non optimisable"
#define MSG_TABLE_READ_ONLY "Les tables %s sont en lecture seulement   "
#define MSG_TABMUL_READONLY "Les tables multiples sont en lecture seulement"
#define MSG_TAB_NOT_LOADED  " (certaines tables n'ont put être chargées)"
#define MSG_TAB_NOT_SPEC    "Table non specifiée"
#define MSG_TB_VW_NOTIN_DB  "Table ou view %s pas dans la base de données"
#define MSG_TDB_NXT_NOT_NUL "Tdb.Next non NULL"
#define MSG_TDB_USE_ERROR   "Erreur, Tdbp->Use=%d"
#define MSG_TOO_MANY_COLS   "Trop de colonnes"
#define MSG_TOO_MANY_COLTAB "Trop de colonnes dans %s (%d)"
#define MSG_TOO_MANY_FIELDS "Trop de champs ligne %d de %s"
#define MSG_TOO_MANY_JUMPS  "Trop de niveaux de saut"
#define MSG_TOO_MANY_KEYS   "Trop de clés (%d)"
#define MSG_TOO_MANY_POS    "Trop de pos_codes"
#define MSG_TOO_MANY_TABLES "Trop de tables (%d)"
#define MSG_TOPSEM_ERROR    "Erreur inconnue dans TopSem"
#define MSG_TO_BLK_IS_NULL  "To Blk est nul"
#define MSG_TO_FTR_NOT_NULL "Set.To_Ftr n'est pas nul"
#define MSG_TO_PIX_NOT_NULL "Set.To_Pix n'est pas nul"
#define MSG_TO_SEM_NOT_NULL "Set.To_Sem n'est pas nul"
#define MSG_TRUNCATE_ERROR  "Erreur en troncation: %s"
#define MSG_TRUNC_BY_ESTIM  "Tronqué par l'option Estimate"
#define MSG_TYPES_ERROR     "Erreur sur Types(%d)"
#define MSG_TYPE_CONV_ERROR "Type non convertible dans une expression"
#define MSG_TYPE_DEF_MISM   "Disparité entre type et définition"
#define MSG_TYPE_MISMATCH   "Clé et source ne sont pas du même type"
#define MSG_TYPE_RECFM_MISM "Disparité entre Type et Recfm"
#define MSG_TYPE_TO_VERIFY  "Type à vérifier: %d"
#define MSG_TYPE_VALUE_ERR  "Colonne %s: disparité type(%s)/valeur(%s)"
#define MSG_UNBALANCE_QUOTE "Appostrophe en trop ligne %d"
#define MSG_UNDEFINED_AM    "COLBLK %s: méthode d'accès indéfinie"
#define MSG_UNDEFINED_PATH  "Chemin d'accès indéfini pour Plgcnx.ini"
#define MSG_UNDEF_COL_COUNT "Count sur colonne non définie"
#define MSG_UNKNOWN_DOMAIN  "Domaine inconnu %s"
#define MSG_UNKNOWN_ERROR   "Erreur inconnue"
#define MSG_UNKNOWN_EXCPT   "Exception non répertoriée"
#define MSG_UNKNOWN_NAME    "Nom inconnu: %.8s"
#define MSG_UNKNOWN_PATH    "Chemin d'accès inconnu pour Plgcnx.ini"
#define MSG_UNKNOWN_POS     "Nom pos_code inconnu: %s"
#define MSG_UNKNOWN_SEM     "Sem %.8s inconnue, rc=%d"
#define MSG_UNKNOWN_SYNONYM "Synonyme inconnu"
#define MSG_UNKNW_QRY_TYPE  "ReadDB: type de requête inconnu"
#define MSG_UNKN_ERR_CODE   "Erreur de code %d inconnu"
#define MSG_UNLOADABLE      " inchargeable: "
#define MSG_UNLOADABLE_PRM  "%s inchargeable: %s"
#define MSG_UNMATCH_FIL_ARG "Argument de filtre dépareillé"
#define MSG_UNQ_COL_SEV_TAB "La colonne %s non qualifiée est dans plusieurs tables"
#define MSG_UNRESOLVED_ARG  "?Argument manquant: %s non résolu en %d ligne %d"
#define MSG_UPDATE_ERROR    "Erreur en Update sur %s"
#define MSG_UPDATING_ROWS   "Mise à jour des lignes"
#define MSG_UPD_ZIP_NOT_IMP "Mise à jour des tables ZDOS non encore implementé"
#define MSG_UP_LANGUAGE     "Bloc langage %.8s version %d niveau %d chargé"
#define MSG_USED_FREE_MEM   "Sarea: utilisé %d, libre %d"
#define MSG_USETEMP_IS      "Usetemp est : %s"
#define MSG_USETEMP_RESET   ". Usetemp remis à Auto"
#define MSG_USETEMP_SET     "Usetemp fixé à %s"
#define MSG_USE_NO_MATCH    "Use non correspondant : Use=%d, ti2=%d, ti3=%d"
#define MSG_USING_INDEX     " (Indexé par"
#define MSG_VALIST_MISMATCH "Disparité des listes de valeurs"
#define MSG_VALSTR_TOO_LONG "Valeur %s trop longue pour une chaîne de longueur %d"
#define MSG_VALTYPE_NOMATCH "Disparité types de valeur"
#define MSG_VALUE_ERROR     "Colonne %s: bloc valeur nul"
#define MSG_VALUE_NOT_ALLOC "Valeur non allouée pour la colonne R%d %s"
#define MSG_VALUE_TOO_BIG   "Valeur %lld trop grande pour la colonne %s"
#define MSG_VALUE_TOO_LONG  "Valeur %s trop longue pour la colonne %s de longueur %d"
#define MSG_VAL_ALLOC_ERR   "Allocation impossible du noeud valeur"
#define MSG_VAL_TOO_LONG    "Valeur %s trop longue pour le champ %s"
#define MSG_VIEW_ALREADY    "La VIEW %s existe déjà"
#define MSG_VIEW_CREATED    "%s view %s créée"
#define MSG_VIEW_DROPPED    "View %s supprimée"
#define MSG_VIEW_NOT_IN_DB  "%s n'est pas une View de %s"
#define MSG_VIR_NO_DELETE   "Delete impossible sur les tables %s"
#define MSG_VIR_READ_ONLY   "Les tables virtuelles %s sont en lecture seulement"
#define MSG_VM_LANG         "Langage au format VM, non supporté"
#define MSG_VOID_FIRST_ARG  "Le premier argument ne doit pas être vide"
#define MSG_VOID_IN_STRING  "Erreur: chaîne IN vide"
#define MSG_VOID_ORDER_LIST "Liste de tri vide, erreur système ?"
#define MSG_VOID_POS_DICT   "Dictionnaire interne du langage vide"
#define MSG_VOID_QUERY      "Requête vide %s"
#define MSG_WORK_AREA       "Espace de travail: %s"
#define MSG_WORK_TOO_SMALL  "Zone de travail trop petite, accroître AreaSize"
#define MSG_WRITE_ERROR     "Erreur à l'écriture de %s"
#define MSG_WRITE_SEEK_ERR  "Erreur de recherche en écriture: %s"
#define MSG_WRITE_STRERROR  "Erreur en écriture sur %s: %s"
#define MSG_WRITING         "Ecriture"
#define MSG_WRITING_ERROR   "Erreur à l'écriture de %s: %s"
#define MSG_WRITING_QUERY   "Erreur à l'écriture de la requête: "
#define MSG_WRONG_ARG_NUM   "La fonction %s ne prend pas %d arguments"
#define MSG_WRONG_COL_NUM   "Numéro de colonne %d trop grand pour %s"
#define MSG_WRONG_DB_LIST   "Liste des bases de données incorrecte ou vide"
#define MSG_WRONG_FUNCTION  "Mauvaise fonction %d"
#define MSG_WRONG_OP_PARM   "Mauvais opérateur ou paramètres pour %s"
#define MSG_WRONG_PARMS     "Mauvais paramètres pour %s"
#define MSG_WRONG_PASSWORD  "Mot de passe illégal pour %s"
#define MSG_WRONG_TYPE      "type non supporté"
#define MSG_WRONG_USERFILE  "La Userfile a une mauvaise taille %d"
#define MSG_WS_CONV_ERR     "Erreur de convertion de %s en WS"
#define MSG_XCOL_MISMATCH   "La colonne %s ne correspond pas à l'index"
#define MSG_XDB_DEL_ERROR   "Erreur en supprimant des entrées du fichier XDB"
#define MSG_XFILE_READERR   "Erreur %d en lisant le fichier index"
#define MSG_XFILE_TOO_SMALL "Le fichier index est plus petit que la taille de l'index"
#define MSG_XFILE_WRITERR   "Erreur en écrivant le fichier index: %s"
#define MSG_XMLTAB_INIT_ERR "Erreur d'initialisation de la table XML"
#define MSG_XML_INIT_ERROR  "Erreur d'initialisation du nouveau fichier XML"
#define MSG_XPATH_CNTX_ERR  "Le nouveau contexte XPath ne peut être créé"
#define MSG_XPATH_EVAL_ERR  "Impossible d'évaluer l'emplacement xpath '%s'"
#define MSG_XPATH_NOT_SUPP  "Xpath non supporté colonne %s"
#define MSG_X_ARG_ADDED     "%d arguments ajoutés"
#define MSG_X_ARG_SET       "%d arguments ont été initialisés"
#define MSG_X_ON_TAB        " %s sur %s("
#define MSG_ZERO_DIVIDE     "Division par zéro dans une expression"
