--disable_warnings
--disable_query_log
--disable_result_log
--source ../t/test_init.inc
--enable_result_log
--enable_query_log
--enable_warnings

--let $MASTER_1_COMMENT_2_1_BACKUP= $MASTER_1_COMMENT_2_1
--let $MASTER_1_COMMENT_2_1_F_BACKUP= $MASTER_1_COMMENT_2_1_F
let $MASTER_1_COMMENT_2_1=
  COMMENT='database "ts_test_remote", table "tbl_a", srv "s_2_1"';
let $MASTER_1_COMMENT_2_1_F=
  COMMENT='database "ts_test_remote", table "tbl_f", srv "s_2_1"';
let $MASTER_1_AUTO_INCREMENT_2_1=
  AUTO_INCREMENT=17 DEFAULT CHARSET=utf8mb4;
let $MASTER_1_AUTO_INCREMENT1=
  AUTO_INCREMENT=17 DEFAULT CHARSET=utf8mb4;
let $MASTER_1_AUTO_INCREMENT2=
  AUTO_INCREMENT=17 DEFAULT CHARSET=utf8mb4;
let $CHILD2_1_AUTO_INCREMENT=
  AUTO_INCREMENT=17 DEFAULT CHARSET=utf8mb4;
--let $CHILD2_1_DROP_TABLE_BACKUP= $CHILD2_1_DROP_TABLE
--let $CHILD2_1_DROP_TABLE_F_BACKUP= $CHILD2_1_DROP_TABLE_F
let $CHILD2_1_DROP_TABLE=
  DROP TABLE IF EXISTS tbl_a;
let $CHILD2_1_DROP_TABLE_F=
  DROP TABLE IF EXISTS tbl_f;
--let $CHILD2_1_CREATE_TABLE_BACKUP= $CHILD2_1_CREATE_TABLE
--let $CHILD2_1_CREATE_TABLE_F_BACKUP= $CHILD2_1_CREATE_TABLE_F
let $CHILD2_1_CREATE_TABLE=
  CREATE TABLE tbl_a (
    col_a INT UNSIGNED NOT NULL AUTO_INCREMENT,
    col_dt DATETIME,
    col_ts TIMESTAMP NOT NULL
      DEFAULT current_timestamp() ON UPDATE current_timestamp(),
    PRIMARY KEY(col_a),
    UNIQUE INDEX i_ts (col_ts)
  ) $CHILD2_1_ENGINE $CHILD2_1_AUTO_INCREMENT;
let $CHILD2_1_CREATE_TABLE_F=
  CREATE TABLE tbl_f (
    col_d DATE,
    col_t TIME
  ) $CHILD2_1_ENGINE;
--let $CHILD2_1_SELECT_TABLE_BACKUP= $CHILD2_1_SELECT_TABLE
--let $CHILD2_1_SELECT_TABLE_F_BACKUP= $CHILD2_1_SELECT_TABLE_F
let $CHILD2_1_SELECT_TABLE=
  SELECT col_a, col_dt, col_ts, unix_timestamp(col_ts) FROM tbl_a ORDER BY col_a;
let $CHILD2_1_SELECT_TABLE_F=
  SELECT col_d, col_t FROM tbl_f;
let $CHILD2_1_SELECT_ARGUMENT1=
  SELECT argument FROM mysql.general_log WHERE argument LIKE '%select %';
let $CHILD2_1_DROP_INDEX=
  DROP INDEX i_ts ON tbl_a;
let $CHILD2_1_SHOW_CREATE_TABLE=
  SHOW CREATE TABLE tbl_a;
let $CHILD2_1_SHOW_CREATE_TABLE_F=
  SHOW CREATE TABLE tbl_f;
--let $OUTPUT_CHILD_GROUP2_BACKUP= $OUTPUT_CHILD_GROUP2
--let $OUTPUT_CHILD_GROUP2= 1
--let $USE_GENERAL_LOG_BACKUP= $USE_GENERAL_LOG
--let $USE_GENERAL_LOG= 1
