LIBRARY		udf_example
VERSION		1.0
EXPORTS
  lookup
  lookup_init
  lookup_deinit
  reverse_lookup
  reverse_lookup_init
  reverse_lookup_deinit
  metaphon_init
  metaphon_deinit
  metaphon
  myfunc_double_init
  myfunc_double
  myfunc_int_init
  myfunc_int
  udf_sequence_init
  udf_sequence_deinit
  udf_sequence
  avgcost_init
  avgcost_deinit
  avgcost_reset
  avgcost_add
  avgcost_clear
  avgcost
  avg2_init
  avg2_deinit
  avg2_reset
  avg2_add
  avg2_remove
  avg2_clear
  avg2
  is_const
  is_const_init
  check_const_len
  check_const_len_init
