let $SLAVE1_1_COMMENT_INCREMENT1_1=
  COMMENT '';
let $SLAVE1_1_COMMENT_INCREMENT1_P_1=
  COMMENT ''
  PARTITION BY LIST(MOD(id, 2)) (
    PARTITION pt1 VALUES IN (0)
    COMMENT='',
    PARTITION pt2 VALUES IN (1)
    COMMENT=''
  );
