#define GRN_EXPR_TOKEN_START_OUTPUT_COLUMNS             1
#define GRN_EXPR_TOKEN_START_ADJUSTER                   2
#define GRN_EXPR_TOKEN_LOGICAL_AND                      3
#define GRN_EXPR_TOKEN_LOGICAL_AND_NOT                  4
#define GRN_EXPR_TOKEN_LOGICAL_OR                       5
#define GRN_EXPR_TOKEN_NEGATIVE                         6
#define GRN_EXPR_TOKEN_QSTRING                          7
#define GRN_EXPR_TOKEN_PARENL                           8
#define GRN_EXPR_TOKEN_PARENR                           9
#define GRN_EXPR_TOKEN_ADJUST                          10
#define GRN_EXPR_TOKEN_RELATIVE_OP                     11
#define GRN_EXPR_TOKEN_IDENTIFIER                      12
#define GRN_EXPR_TOKEN_BRACEL                          13
#define GRN_EXPR_TOKEN_BRACER                          14
#define GRN_EXPR_TOKEN_EVAL                            15
#define GRN_EXPR_TOKEN_COMMA                           16
#define GRN_EXPR_TOKEN_ASSIGN                          17
#define GRN_EXPR_TOKEN_STAR_ASSIGN                     18
#define GRN_EXPR_TOKEN_SLASH_ASSIGN                    19
#define GRN_EXPR_TOKEN_MOD_ASSIGN                      20
#define GRN_EXPR_TOKEN_PLUS_ASSIGN                     21
#define GRN_EXPR_TOKEN_MINUS_ASSIGN                    22
#define GRN_EXPR_TOKEN_SHIFTL_ASSIGN                   23
#define GRN_EXPR_TOKEN_SHIFTR_ASSIGN                   24
#define GRN_EXPR_TOKEN_SHIFTRR_ASSIGN                  25
#define GRN_EXPR_TOKEN_AND_ASSIGN                      26
#define GRN_EXPR_TOKEN_XOR_ASSIGN                      27
#define GRN_EXPR_TOKEN_OR_ASSIGN                       28
#define GRN_EXPR_TOKEN_QUESTION                        29
#define GRN_EXPR_TOKEN_COLON                           30
#define GRN_EXPR_TOKEN_BITWISE_OR                      31
#define GRN_EXPR_TOKEN_BITWISE_XOR                     32
#define GRN_EXPR_TOKEN_BITWISE_AND                     33
#define GRN_EXPR_TOKEN_EQUAL                           34
#define GRN_EXPR_TOKEN_NOT_EQUAL                       35
#define GRN_EXPR_TOKEN_LESS                            36
#define GRN_EXPR_TOKEN_GREATER                         37
#define GRN_EXPR_TOKEN_LESS_EQUAL                      38
#define GRN_EXPR_TOKEN_GREATER_EQUAL                   39
#define GRN_EXPR_TOKEN_IN                              40
#define GRN_EXPR_TOKEN_MATCH                           41
#define GRN_EXPR_TOKEN_NEAR                            42
#define GRN_EXPR_TOKEN_NEAR2                           43
#define GRN_EXPR_TOKEN_SIMILAR                         44
#define GRN_EXPR_TOKEN_TERM_EXTRACT                    45
#define GRN_EXPR_TOKEN_LCP                             46
#define GRN_EXPR_TOKEN_PREFIX                          47
#define GRN_EXPR_TOKEN_SUFFIX                          48
#define GRN_EXPR_TOKEN_REGEXP                          49
#define GRN_EXPR_TOKEN_SHIFTL                          50
#define GRN_EXPR_TOKEN_SHIFTR                          51
#define GRN_EXPR_TOKEN_SHIFTRR                         52
#define GRN_EXPR_TOKEN_PLUS                            53
#define GRN_EXPR_TOKEN_MINUS                           54
#define GRN_EXPR_TOKEN_STAR                            55
#define GRN_EXPR_TOKEN_SLASH                           56
#define GRN_EXPR_TOKEN_MOD                             57
#define GRN_EXPR_TOKEN_DELETE                          58
#define GRN_EXPR_TOKEN_INCR                            59
#define GRN_EXPR_TOKEN_DECR                            60
#define GRN_EXPR_TOKEN_NOT                             61
#define GRN_EXPR_TOKEN_BITWISE_NOT                     62
#define GRN_EXPR_TOKEN_EXACT                           63
#define GRN_EXPR_TOKEN_PARTIAL                         64
#define GRN_EXPR_TOKEN_UNSPLIT                         65
#define GRN_EXPR_TOKEN_DECIMAL                         66
#define GRN_EXPR_TOKEN_HEX_INTEGER                     67
#define GRN_EXPR_TOKEN_STRING                          68
#define GRN_EXPR_TOKEN_BOOLEAN                         69
#define GRN_EXPR_TOKEN_NULL                            70
#define GRN_EXPR_TOKEN_BRACKETL                        71
#define GRN_EXPR_TOKEN_BRACKETR                        72
#define GRN_EXPR_TOKEN_DOT                             73
#define GRN_EXPR_TOKEN_NONEXISTENT_COLUMN              74
