/* Copyright (C) MariaDB Corporation Ab */
#define MSG_ACCESS_VIOLATN     200
#define MSG_ADD_BAD_TYPE       201
#define MSG_ALLOC_ERROR        202
#define MSG_ANSWER_TYPE        203
#define MSG_API_CONF_ERROR     204
#define MSG_APPL_NOT_INIT      205
#define MSG_ARRAY_BNDS_EXCD    206
#define MSG_BAD_ARRAY_OPER     207
#define MSG_BAD_ARRAY_TYPE     208
#define MSG_BAD_ARRAY_VAL      209
#define MSG_BAD_BIN_FMT        210
#define MSG_BAD_BLK_ESTIM      211
#define MSG_BAD_BLK_SIZE       212
#define MSG_BAD_BYTE_NUM       213
#define MSG_BAD_BYTE_READ      214
#define MSG_BAD_COL_TYPE       215
#define MSG_BAD_COL_XPATH      216
#define MSG_BAD_CONST_TYPE     217
#define MSG_BAD_CONV_TYPE      218
#define MSG_BAD_DATETIME       219
#define MSG_BAD_DBF_FILE       220
#define MSG_BAD_DBF_REC        221
#define MSG_BAD_DBF_TYPE       222
#define MSG_BAD_DIRECTORY      223
#define MSG_BAD_FIELD_RANK     224
#define MSG_BAD_FIELD_TYPE     225
#define MSG_BAD_FILE_HANDLE    226
#define MSG_BAD_FILTER         227
#define MSG_BAD_FILTER_CONV    228
#define MSG_BAD_FILTER_OP      229
#define MSG_BAD_FLD_FORMAT     230
#define MSG_BAD_FLD_LENGTH     231
#define MSG_BAD_FREQ_SET       232
#define MSG_BAD_FUNC_MODE      233
#define MSG_BAD_HANDLE_VAL     234
#define MSG_BAD_HEADER         235
#define MSG_BAD_HEAD_END       236
#define MSG_BAD_INDEX_FILE     237
#define MSG_BAD_LINEFLD_FMT    238
#define MSG_BAD_LINE_LEN       239
#define MSG_BAD_LRECL          240
#define MSG_BAD_NODE_TYPE      241
#define MSG_BAD_OFFSET_VAL     242
#define MSG_BAD_OPEN_MODE      243
#define MSG_BAD_PARAM_TYPE     244
#define MSG_BAD_PARM_COUNT     245
#define MSG_BAD_QUOTE_FIELD    246
#define MSG_BAD_READ_NUMBER    247
#define MSG_BAD_RECFM          248
#define MSG_BAD_RECFM_VAL      249
#define MSG_BAD_SET_CASE       250
#define MSG_BAD_SET_STRING     251
#define MSG_BAD_SPECIAL_COL    252
#define MSG_BAD_SPEC_COLUMN    253
#define MSG_BAD_TABLE_TYPE     254
#define MSG_BAD_TYPE_LIKE      255
#define MSG_BAD_VALBLK_INDX    256
#define MSG_BAD_VALBLK_TYPE    257
#define MSG_BAD_VALNODE        258
#define MSG_BAD_VALUE_TYPE     259
#define MSG_BAD_VAL_UPDATE     260
#define MSG_BAS_NS_LIST        261
#define MSG_BIN_F_TOO_LONG     262
#define MSG_BIN_MODE_FAIL      263
#define MSG_BLKTYPLEN_MISM     264
#define MSG_BLK_IS_NULL        265
#define MSG_BREAKPOINT         266
#define MSG_BUILD_INDEX        267
#define MSG_CANNOT_OPEN        268
#define MSG_CHSIZE_ERROR       269
#define MSG_COL_ALLOC_ERR      270
#define MSG_COL_ISNOT_TABLE    271
#define MSG_COL_NOT_SORTED     272
#define MSG_COL_NUM_MISM       273
#define MSG_COM_ERROR          274
#define MSG_CONCAT_SUBNODE     275
#define MSG_CONNECT_CANCEL     276
#define MSG_CONTROL_C_EXIT     277
#define MSG_DATABASE_LOADED    278
#define MSG_DATA_MISALIGN      279
#define MSG_DBASE_FILE         280
#define MSG_DEF_ALLOC_ERROR    281
#define MSG_DEL_FILE_ERR       282
#define MSG_DEL_READ_ERROR     283
#define MSG_DEL_WRITE_ERROR    284
#define MSG_DEPREC_FLAG        285
#define MSG_DLL_LOAD_ERROR     286
#define MSG_DOM_NOT_SUPP       287
#define MSG_DVAL_NOTIN_LIST    288
#define MSG_EMPTY_DOC          289
#define MSG_EMPTY_FILE         290
#define MSG_EOF_AFTER_LINE     291
#define MSG_EOF_INDEX_FILE     292
#define MSG_ERROR_IN_LSK       293
#define MSG_ERROR_IN_SFP       294
#define MSG_ERR_READING_REC    295
#define MSG_FAIL_ADD_NODE      296
#define MSG_FETCH_NO_RES       297
#define MSG_FIELD_TOO_LONG     298
#define MSG_FILELEN_ERROR      299
#define MSG_FILE_IS_EMPTY      300
#define MSG_FILE_MAP_ERR       301
#define MSG_FILE_MAP_ERROR     302
#define MSG_FILE_OPEN_YET      303
#define MSG_FILE_UNFOUND       304
#define MSG_FLD_TOO_LNG_FOR    305
#define MSG_FLT_BAD_RESULT     306
#define MSG_FLT_DENORMAL_OP    307
#define MSG_FLT_INVALID_OP     308
#define MSG_FLT_OVERFLOW       309
#define MSG_FLT_STACK_CHECK    310
#define MSG_FLT_UNDERFLOW      311
#define MSG_FLT_ZERO_DIVIDE    312
#define MSG_FMT_WRITE_NIY      313
#define MSG_FOXPRO_FILE        314
#define MSG_FPUTS_ERROR        315
#define MSG_FSEEK_ERROR        316
#define MSG_FSETPOS_ERROR      317
#define MSG_FTELL_ERROR        318
#define MSG_FUNCTION_ERROR     319
#define MSG_FUNC_ERRNO         320
#define MSG_FUNC_ERROR         321
#define MSG_FUNC_ERR_S         322
#define MSG_FWRITE_ERROR       323
#define MSG_GET_DIST_VALS      324
#define MSG_GET_FUNC_ERR       325
#define MSG_GLOBAL_ERROR       326
#define MSG_GUARD_PAGE         327
#define MSG_GZOPEN_ERROR       328
#define MSG_ILLEGAL_INSTR      329
#define MSG_ILL_FILTER_CONV    330
#define MSG_INDEX_NOT_UNIQ     331
#define MSG_INDEX_YET_ON       332
#define MSG_INDX_COL_NOTIN     333
#define MSG_INDX_EXIST_YET     334
#define MSG_INIT_FAILED        335
#define MSG_INT_COL_ERROR      336
#define MSG_INT_OVERFLOW       337
#define MSG_INT_ZERO_DIVIDE    338
#define MSG_INVALID_DISP       339
#define MSG_INVALID_FTYPE      340
#define MSG_INVALID_HANDLE     341
#define MSG_INVALID_OPER       342
#define MSG_INV_COLUMN_TYPE    343
#define MSG_INV_COL_TYPE       344
#define MSG_INV_DEF_READ       345
#define MSG_INV_DIRCOL_OFST    346
#define MSG_INV_MAP_POS        347
#define MSG_INV_RAND_ACC       348
#define MSG_INV_REC_POS        349
#define MSG_INV_RESULT_TYPE    350
#define MSG_INV_UPDT_TABLE     351
#define MSG_IN_WITHOUT_SUB     352
#define MSG_KEY_ALLOC_ERR      353
#define MSG_KEY_ALLOC_ERROR    354
#define MSG_LINE_TOO_LONG      355
#define MSG_LIST               356
#define MSG_LOADING_FAILED     357
#define MSG_LRECL_TOO_SMALL    358
#define MSG_MAKE_EMPTY_FILE    359
#define MSG_MAKING             360
#define MSG_MALLOC_ERROR       361
#define MSG_MAP_VIEW_ERROR     362
#define MSG_MAXSIZE_ERROR      363
#define MSG_MEM_ALLOC_ERR      364
#define MSG_MEM_ALLOC_ERROR    365
#define MSG_MISPLACED_QUOTE    366
#define MSG_MISSING_ARG        367
#define MSG_MISSING_FIELD      368
#define MSG_MISSING_FNAME      369
#define MSG_MISSING_NODE       370
#define MSG_MISSING_ROWNODE    371
#define MSG_MIS_TAG_LIST       372
#define MSG_MUL_MAKECOL_ERR    373
#define MSG_NAME_CONV_ERR      374
#define MSG_NEW_DOC_FAILED     375
#define MSG_NEW_RETURN_NULL    376
#define MSG_NEXT_FILE_ERROR    377
#define MSG_NONCONT_EXCEPT     378
#define MSG_NOP_ZLIB_INDEX     379
#define MSG_NOT_A_DBF_FILE     380
#define MSG_NOT_FIXED_LEN      381
#define MSG_NO_0DH_HEAD        382
#define MSG_NO_ACTIVE_DB       383
#define MSG_NO_CHAR_FROM       384
#define MSG_NO_DATE_FMT        385
#define MSG_NO_DEF_FNCCOL      386
#define MSG_NO_DEF_PIVOTCOL    387
#define MSG_NO_DIR_INDX_RD     388
#define MSG_NO_FEAT_SUPPORT    389
#define MSG_NO_FLD_FORMAT      390
#define MSG_NO_FORMAT_COL      391
#define MSG_NO_FORMAT_TYPE     392
#define MSG_NO_INDEX_READ      393
#define MSG_NO_KEY_COL         394
#define MSG_NO_KEY_UPDATE      395
#define MSG_NO_MAP_INSERT      396
#define MSG_NO_MATCHING_COL    397
#define MSG_NO_MATCH_COL       398
#define MSG_NO_MEMORY          399
#define MSG_NO_MODE_PADDED     400
#define MSG_NO_MUL_VCT         401
#define MSG_NO_ODBC_DELETE     402
#define MSG_NO_ODBC_DIRECT     403
#define MSG_NO_ODBC_MUL        404
#define MSG_NO_ODBC_SPECOL     405
#define MSG_NO_PART_DEL        406
#define MSG_NO_PART_MAP        407
#define MSG_NO_PAR_BLK_INS     408
#define MSG_NO_PIV_DIR_ACC     409
#define MSG_NO_READ_32         410
#define MSG_NO_RECOV_SPACE     411
#define MSG_NO_ROWID_FOR_AM    412
#define MSG_NO_ROW_NODE        413
#define MSG_NO_SECTION_NAME    414
#define MSG_NO_SEC_UPDATE      415
#define MSG_NO_SETPOS_YET      416
#define MSG_NO_SPEC_COL        417
#define MSG_NO_SUB_VAL         418
#define MSG_NO_TABCOL_DATA     419
#define MSG_NO_TABLE_DEL       420
#define MSG_NO_TAB_DATA        421
#define MSG_NO_VCT_DELETE      422
#define MSG_NO_ZIP_DELETE      423
#define MSG_OPENING            424
#define MSG_OPEN_EMPTY_FILE    425
#define MSG_OPEN_ERROR         426
#define MSG_OPEN_ERROR_IS      427
#define MSG_OPEN_MODE_ERROR    428
#define MSG_OPEN_STRERROR      429
#define MSG_OPTBLK_RD_ERR      430
#define MSG_OPTBLK_WR_ERR      431
#define MSG_OPTIMIZING         432
#define MSG_OPT_BMAP_RD_ERR    433
#define MSG_OPT_BMAP_WR_ERR    434
#define MSG_OPT_CANCELLED      435
#define MSG_OPT_DVAL_RD_ERR    436
#define MSG_OPT_DVAL_WR_ERR    437
#define MSG_OPT_HEAD_RD_ERR    438
#define MSG_OPT_HEAD_WR_ERR    439
#define MSG_OPT_LOGIC_ERR      440
#define MSG_OPT_MAX_RD_ERR     441
#define MSG_OPT_MAX_WR_ERR     442
#define MSG_OPT_MIN_RD_ERR     443
#define MSG_OPT_MIN_WR_ERR     444
#define MSG_OPT_NOT_MATCH      445
#define MSG_PAGE_ERROR         446
#define MSG_PARM_CNT_MISS      447
#define MSG_PREC_VBLP_NULL     448
#define MSG_PRIV_INSTR         449
#define MSG_PROCADD_ERROR      450
#define MSG_QUERY_CANCELLED    451
#define MSG_RANGE_NO_JOIN      452
#define MSG_RC_READING         453
#define MSG_READY              454
#define MSG_READ_ERROR         455
#define MSG_READ_ONLY          456
#define MSG_READ_SEEK_ERROR    457
#define MSG_REGISTER_ERR       458
#define MSG_REMOVE_ERROR       459
#define MSG_RENAME_ERROR       460
#define MSG_ROWID_NOT_IMPL     461
#define MSG_SEC_KEY_FIRST      462
#define MSG_SEC_NAME_FIRST     463
#define MSG_SEP_IN_FIELD       464
#define MSG_SEQUENCE_ERROR     465
#define MSG_SETEOF_ERROR       466
#define MSG_SETRECPOS_NIY      467
#define MSG_SET_STR_TRUNC      468
#define MSG_SFP_ERROR          469
#define MSG_SHARED_LIB_ERR     470
#define MSG_SINGLE_STEP        471
#define MSG_SORTING_VAL        472
#define MSG_SPCOL_READONLY     473
#define MSG_SQL_CONF_ERROR     474
#define MSG_SRCH_CLOSE_ERR     475
#define MSG_SRC_TABLE_UNDEF    476
#define MSG_STACK_OVERFLOW     477
#define MSG_TABDIR_READONLY    478
#define MSG_TABLE_NOT_OPT      479
#define MSG_TABLE_NO_INDEX     480
#define MSG_TABLE_READ_ONLY    481
#define MSG_TABMUL_READONLY    482
#define MSG_TOO_MANY_FIELDS    483
#define MSG_TOO_MANY_JUMPS     484
#define MSG_TOO_MANY_KEYS      485
#define MSG_TO_BLK_IS_NULL     486
#define MSG_TRUNCATE_ERROR     487
#define MSG_TRUNC_BY_ESTIM     488
#define MSG_TYPE_MISMATCH      489
#define MSG_TYPE_VALUE_ERR     490
#define MSG_UNBALANCE_QUOTE    491
#define MSG_UNDEFINED_AM       492
#define MSG_UNKNOWN_EXCPT      493
#define MSG_UNMATCH_FIL_ARG    494
#define MSG_UPDATE_ERROR       495
#define MSG_UPD_ZIP_NOT_IMP    496
#define MSG_VALSTR_TOO_LONG    497
#define MSG_VALTYPE_NOMATCH    498
#define MSG_VALUE_ERROR        499
#define MSG_VALUE_TOO_BIG      500
#define MSG_VALUE_TOO_LONG     501
#define MSG_VAL_ALLOC_ERR      502
#define MSG_VIR_NO_DELETE      503
#define MSG_VIR_READ_ONLY      504
#define MSG_VOID_FIRST_ARG     505
#define MSG_WORK_AREA          506
#define MSG_WRITE_SEEK_ERR     507
#define MSG_WRITE_STRERROR     508
#define MSG_WRITING            509
#define MSG_WRITING_ERROR      510
#define MSG_WS_CONV_ERR        511
#define MSG_XCOL_MISMATCH      512
#define MSG_XFILE_READERR      513
#define MSG_XFILE_WRITERR      514
#define MSG_XMLTAB_INIT_ERR    515
#define MSG_XML_INIT_ERROR     516
#define MSG_XPATH_CNTX_ERR     517
#define MSG_XPATH_EVAL_ERR     518
#define MSG_XPATH_NOT_SUPP     519
#define MSG_ZERO_DIVIDE        520
#define MSG_FIX_OVFLW_ADD      521
#define MSG_FIX_OVFLW_TIMES    522
#define MSG_FIX_UNFLW_ADD      523
#define MSG_FIX_UNFLW_TIMES    524
